# 20210609
import numpy as np
import scipy.signal

from PyQt5 import QtCore, QtWidgets, QtGui

import matplotlib.pyplot as plt

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

import sanpy
from sanpy import bAnalysis
from sanpy.interface.plugins import sanpyPlugin
from sanpy.bAnalysisUtil import statList
from sanpy.interface.plugins import ResponseType  # to toggle response to set sweeps, etc

class plotRecording(sanpyPlugin):
    """
    Example of matplotlib plugin.
    """
    myHumanName = 'Plot Recording'

    def __init__(self, **kwargs):
        super(plotRecording, self).__init__(**kwargs)

        # this is a very simple plugin, do not respond to changes in interface
        # switchFile = self.responseTypes.switchFile
        # self.toggleResponseOptions(switchFile, newValue=False)

        # analysisChange = self.responseTypes.analysisChange
        # self.toggleResponseOptions(analysisChange, newValue=False)

        #selectSpike = self.responseTypes.selectSpike
        #self.toggleResponseOptions(selectSpike, newValue=False)

        # we are plotting all sweeps and all axis, turn these off
        self.toggleResponseOptions(ResponseType.setSweep, False)  # we plot all sweeps
        self.toggleResponseOptions(ResponseType.setAxis, False)

        setAxis = self.responseTypes.setAxis
        self.toggleResponseOptions(setAxis, newValue=False)

        self.rawLine = None
        self.thresholdLine = None

        self.plot()
        self.replot()

    def plot(self):
        # if self.ba is None:
        #     return

        self.mplWindow2() # assigns (self.fig, self.axs)

        return

    def replot(self):
        """bAnalysis has been updated, replot
        """
        logger.info(f'{self.ba}')

        if self.ba is None:
            return
            
        self.xOffset = 0.01
        self.yOffset = 50
        
        xCurrentOffset = 0
        yCurrentOffset = 0

        currentSweep = self.ba.fileLoader.currentSweep
        numSweeps = self.ba.fileLoader.numSweeps

        self.fig.clear(True)
        self.axs = self.fig.add_subplot(1,1,1)

        '''
        if self.rawLine is not None:
            for line in self.rawLine:
                line.clear()
        self.rawLine = [None] * numSweeps
        
        if self.thresholdLine is not None:
            for line in self.thresholdLine:
                line.clear()
        self.thresholdLine = [None] * numSweeps
        '''
        self.rawLine = [None] * numSweeps
        self.thresholdLine = [None] * numSweeps

        _penColor = self.getPenColor()

        for sweepIdx in range(numSweeps):
            self.ba.fileLoader.setSweep(sweepIdx)
            
            #self.axs[_idx].self.fig.add_subplot(numSweeps, 1, _idx)

            sweepX = self.getSweep('x')
            sweepY = self.getSweep('y')

            self.sweepX = sweepX
            self.sweepY = sweepY

            self.rawLine[sweepIdx], = self.axs.plot(sweepX+xCurrentOffset, sweepY+yCurrentOffset,
                                                    '-', color=_penColor, linewidth=0.5)

            thresholdSec = self.ba.getStat('thresholdSec', sweepNumber=sweepIdx)
            thresholdSec = [x+xCurrentOffset for x in thresholdSec]
            thresholdVal = self.ba.getStat('thresholdVal', sweepNumber=sweepIdx)
            thresholdVal = [x+yCurrentOffset for x in thresholdVal]
            markersize = 4
            self.thresholdLine[sweepIdx] = self.axs.plot(thresholdSec, thresholdVal, 'r.', markersize=markersize)
            '''
            if thresholdSec is None or thresholdVal is None:
                self.thresholdLine[sweepIdx].set_data([], [])
            else:
                self.thresholdLine[sweepIdx] = self.axs.plot(thresholdSec, thresholdSec, 'o')
            '''

            '''
            xHW, yHW = self.getHalfWidths()
            self.lineHW, = self.axs.plot(xHW, yHW, '-')
            '''

            xCurrentOffset += self.xOffset
            yCurrentOffset += self.yOffset

        self.axs.relim()
        self.axs.autoscale_view(True,True,True)

        self.ba.fileLoader.setSweep(currentSweep)
        self.static_canvas.draw_idle()
        plt.draw()

    def old_getEddLines(self):
        preLinearFitPnt0 = self.ba.getStat('preLinearFitPnt0')
        preLinearFitSec0 = [self.ba.fileLoader.pnt2Sec_(x) for x in preLinearFitPnt0]
        preLinearFitVal0 = self.ba.fileLoader.getStat('preLinearFitVal0')

        preLinearFitPnt1 = self.ba.getStat('preLinearFitPnt1')
        preLinearFitSec1 = [self.ba.fileLoader.pnt2Sec_(x) for x in preLinearFitPnt1]
        preLinearFitVal1 = self.ba.getStat('preLinearFitVal1')

        x = []
        y = []
        for idx in range(self.ba.numSpikes):
            try:
                dx = preLinearFitSec1[idx] - preLinearFitSec0[idx]
                dy = preLinearFitVal1[idx] - preLinearFitVal0[idx]
            except (IndexError) as e:
                logger.error(f'spike {idx} preLinearFitSec1:{len(preLinearFitSec1)} preLinearFitSec0:{len(preLinearFitSec0)}')
                logger.error(f'spike {idx} preLinearFitPnt1:{len(preLinearFitPnt1)} preLinearFitPnt0:{len(preLinearFitPnt0)}')

            lineLength = 4  # TODO: make this a function of spike frequency?

            try:
                x.append(preLinearFitSec0[idx])
                x.append(preLinearFitSec1[idx] + lineLength*dx)
                x.append(np.nan)

                y.append(preLinearFitVal0[idx])
                y.append(preLinearFitVal1[idx] + lineLength*dy)
                y.append(np.nan)
            except (IndexError) as e:
                logger.error(f'preLinearFitSec0:{len(preLinearFitSec0)} preLinearFitSec1:{len(preLinearFitSec1)}')
                logger.error(e)

        return x, y

    def getHalfWidths(self):
        """Get x/y pair for plotting all half widths.
        
        DOes not work with new version because of sweeps.
        """
        # defer until we know how many half-widths 20/50/80
        x = []
        y = []
        numPerSpike = 3  # rise/fall/nan
        numSpikes = self.ba.numSpikes
        xyIdx = 0
        for idx, spike in enumerate(self.ba.spikeDict):
            if idx ==0:
                # make x/y from first spike using halfHeights = [20,50,80]
                halfHeights = spike['halfHeights'] # will be same for all spike, like [20, 50, 80]
                numHalfHeights = len(halfHeights)
                # *numHalfHeights to account for rise/fall + padding nan
                x = [np.nan] * (numSpikes * numHalfHeights * numPerSpike)
                y = [np.nan] * (numSpikes * numHalfHeights * numPerSpike)
                #print('  len(x):', len(x), 'numHalfHeights:', numHalfHeights, 'numSpikes:', numSpikes, 'halfHeights:', halfHeights)

            for idx2, width in enumerate(spike['widths']):
                halfHeight = width['halfHeight'] # [20,50,80]
                risingPnt = width['risingPnt']
                #risingVal = width['risingVal']
                risingVal = self.sweepY[risingPnt]
                fallingPnt = width['fallingPnt']
                #fallingVal = width['fallingVal']
                fallingVal = self.sweepY[fallingPnt]

                if risingPnt is None or fallingPnt is None:
                    # half-height was not detected
                    continue

                risingSec = self.ba.fileLoader.pnt2Sec_(risingPnt)
                fallingSec = self.ba.fileLoader.pnt2Sec_(fallingPnt)

                x[xyIdx] = risingSec
                x[xyIdx+1] = fallingSec
                x[xyIdx+2] = np.nan
                # y
                y[xyIdx] = fallingVal  #risingVal, to make line horizontal
                y[xyIdx+1] = fallingVal
                y[xyIdx+2] = np.nan

                # each spike has 3x pnts: rise/fall/nan
                xyIdx += numPerSpike  # accounts for rising/falling/nan
            # end for width
        # end for spike
        return x, y

    def selectSpikeList(self):
        """Only respond to single spike selection.
        """
        logger.info('')
        spikeList = self.getSelectedSpikes()

        if spikeList==[] or len(spikeList)>1:
            return

        if self.ba is None:
            return

        spikeNumber = spikeList[0]

        thresholdSec = self.ba.getStat('thresholdSec')
        spikeTime = thresholdSec[spikeNumber]
        xMin = spikeTime - 0.5
        xMax = spikeTime + 0.5

        self.axs.set_xlim(xMin, xMax)

        self.fig.canvas.draw()
        self.fig.canvas.flush_events()

    def old_slot_selectSpike(self, eDict):
        logger.info(eDict)
        spikeNumber = eDict['spikeNumber']
        #doZoom = eDict['doZoom']

        if spikeNumber is None:
            return

        if self.ba is None:
            return

        thresholdSec = self.ba.getStat('thresholdSec')
        spikeTime = thresholdSec[spikeNumber]
        xMin = spikeTime - 0.5
        xMax = spikeTime + 0.5

        self.axs.set_xlim(xMin, xMax)

        self.fig.canvas.draw()
        self.fig.canvas.flush_events()

def testPlot():
    import os
    file_path = os.path.realpath(__file__)  # full path to this file
    file_path = os.path.split(file_path)[0]  # folder for this file
    path = os.path.join(file_path, '../../../data/19114001.abf')

    ba = bAnalysis(path)
    if ba.loadError:
        print('error loading file')
        return
    ba.spikeDetect()

    # create plugin
    ap = plotRecording(ba=ba)

    #ap.plot()

    #ap.slotUpdateAnalysis()

def main():
    path = '/Users/cudmore/Sites/SanPy/data/2021_07_20_0010.abf'
    ba = sanpy.bAnalysis(path)
    ba.spikeDetect()
    print(ba.numSpikes)

    import sys
    app = QtWidgets.QApplication([])
    pr = plotRecording(ba=ba)
    pr.show()
    sys.exit(app.exec_())

def testLoad():
    import os, glob
    pluginFolder = '/Users/cudmore/sanpy_plugins'
    files = glob.glob(os.path.join(pluginFolder, '*.py'))
    for file in files:
        #if file.startswith('.'):
        #    continue
        if file.endswith('__init__.py'):
            continue
        print(file)
if __name__ == '__main__':
    #testPlot()
    #testLoad()
    main()



# Author: Robert Cudmore
# Date: 20190627
# todo: move these to a json file !!!
# A list of human readable stats and how to map them to backend
# Each key, like 'Take Off Potential (mV)' is a y-stat

import sys, os, json
from collections import OrderedDict
from typing import List

import sanpy

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

"""
A dictionary of detection parameters.

Functions to map human readable name to backend stat name and vica versa.

See [xxx-methods](../../Methods) section for a list of available stats

Example:

```
from sanpy.bAnalysisUtil import getHumandFromStat, getStatFromHuman

humanStr = getHumandFromStat('thresholdSec')
statStr = getStatFromHuman('Spike Time (s)')
```
"""

# add this to keep all key VALUES the same
# or maybe make a class??? like bDetection (which holds detection parameters?)
'''
def _defaultStatDict():
    return
'''

statList = OrderedDict()
statList['Spike Time (s)'] = {
    'name': 'thresholdSec',
    'units': 's',
    'yStat': 'thresholdVal',
    'yStatUnits': 'mV',
    'xStat': 'thresholdSec',
    'xStatUnits': 's'
    }
statList['Spike Number'] = {
    'name': 'spikeNumber',
    'units': '',
    'yStat': '',
    'yStatUnits': '',
    'xStat': 'spikeNumber',
    'xStatUnits': ''
    }
statList['Sweep Spike Number'] = {
    'name': 'sweepSpikeNumber',
    'units': '',
    'yStat': '',
    'yStatUnits': '',
    'xStat': 'sweepSpikeNumber',
    'xStatUnits': ''
    }
statList['Sweep Number'] = {
    'name': 'sweep',
    'units': '',
    'yStat': 'sweep',
    'yStatUnits': '',
    'xStat': '',
    'xStatUnits': ''
    }
# removed 20230320
# statList['DAC Command'] = {
    # 'name': 'dacCommand',
    # 'units': '',
    # 'yStat': 'dacCommand',
    # 'yStatUnits': '',
    # 'xStat': '',
    # 'xStatUnits': ''
    # }

# added 20230320
# stimulus epoch number a spike is within
statList['Epoch'] = {
    'name': 'epoch',
    'units': '',
    'yStat': 'epoch',
    'yStatUnits': '',
    'xStat': '',
    'xStatUnits': ''
    }
# added 20230320
# the epoch level (for I-Clamp, corresponds to DAC command)
statList['Epoch DAC'] = {
    'name': 'epochLevel',
    'units': '',
    'yStat': 'epochLevel',
    'yStatUnits': '',
    'xStat': '',
    'xStatUnits': ''
    }
# added 20230320
# the spike number within the epoch
# not implemented
statList['Epoch Spike Number'] = {
    'name': 'epochSpikeNumber',
    'units': '',
    'yStat': 'epochSpikeNumber',
    'yStatUnits': '',
    'xStat': '',
    'xStatUnits': ''
    }

statList['Take Off Potential (mV)'] = {
    'name': 'thresholdVal',
    'units': 'mV',
    'yStat': 'thresholdVal',
    'yStatUnits': 'mV',
    'xStat': 'thresholdPnt',
    'xStatUnits': 'Points'
    }
statList['Spike Frequency (Hz)'] = {
    'name': 'spikeFreq_hz',
    'units': 'Hz',
    'yStat': 'spikeFreq_hz',
    'yStatUnits': 'Hz',
    'xStat': 'thresholdPnt',
    'xStatUnits': 'Points'
    }
statList['Inter-Spike-Interval (ms)'] = {
    'name': 'isi_ms',
    'units': 'ms',
    'yStat': 'isi_ms',
    'yStatUnits': 'ms',
    'xStat': 'thresholdPnt',
    'xStatUnits': 'Points'
    }
statList['Cycle Length (ms)'] = {
    'name': 'cycleLength_ms',
    'units': 'ms',
    'yStat': 'cycleLength_ms',
    'yStatUnits': 'ms',
    'xStat': 'thresholdPnt',
    'xStatUnits': 'Points'
    }
statList['AP Peak (mV)'] = {
    'name': 'peakVal',
    'units': 'mV',
    'yStat': 'peakVal',
    'yStatUnits': 'mV',
    'xStat': 'peakPnt',
    'xStatUnits': 'Points'
    }
#spikeDict['peakVal'] - spikeDict['thresholdVal']
statList['AP Height (mV)'] = {
    'name': 'peakHeight',
    'units': 'mV',
    'yStat': 'peakHeight',
    'yStatUnits': 'mV',
    'xStat': 'peakPnt',
    'xStatUnits': 'Points'
    }
# new 20220122
statList['Time To Peak (ms)'] = {
    'name': 'timeToPeak_ms',
    'units': 'ms',
    'yStat': 'peakVal',
    'yStatUnits': 'mV',
    'xStat': 'peakPnt',
    'xStatUnits': 'Points'
    }
statList['Pre AP Min (mV)'] = {
    'name': 'preMinVal',
    'units': 'mV',
    'yStat': 'preMinVal',
    'yStatUnits': 'mV',
    'xStat': 'preMinPnt',
    'xStatUnits': 'Points'
    }
statList['Post AP Min (mV)'] = {
    'name': 'postMinVal',
    'units': 'mV',
    'yStat': 'postMinVal',
    'yStatUnits': 'mV',
    'xStat': 'postMinPnt',
    'xStatUnits': 'Points'
    }
# todo: fix this
statList['Early Diastolic Depol Rate (dV/s)'] = {
    'name': 'earlyDiastolicDurationRate',
    'units': 'dV/s',
    'yStat': 'earlyDiastolicDurationRate',
    'yStatUnits': 'dV/s',
    'xStat': '',
    'xStatUnits': ''
    }
# todo: fix this
statList['Early Diastolic Duration (ms)'] = {
    'name': 'earlyDiastolicDuration_ms',
    'units': 'ms',
    'yStat': 'earlyDiastolicDuration_ms',
    'yStatUnits': 'dV/s',
    'xStat': 'thresholdPnt',
    'xStatUnits': 'Points'
    }

statList['Diastolic Duration (ms)'] = {
    'name': 'diastolicDuration_ms',
    'units': 'ms',
    'yStat': 'diastolicDuration_ms',
    'yStatUnits': 'dV/s',
    'xStat': 'thresholdPnt',
    'xStatUnits': 'Points'
    }
statList['Max AP Upstroke (mV)'] = {
    'name': 'preSpike_dvdt_max_val',
    'units': 'mV',
    'yStat': 'preSpike_dvdt_max_val',
    'yStatUnits': 'dV/s',
    'xStat': 'preSpike_dvdt_max_pnt',
    'xStatUnits': 'Points'
    }
statList['Max AP Upstroke (dV/dt)'] = {
    'name': 'preSpike_dvdt_max_val2',
    'units': 'dV/dt',
    'yStat': 'preSpike_dvdt_max_val2',
    'yStatUnits': 'dV/dt',
    'xStat': 'preSpike_dvdt_max_pnt',
    'xStatUnits': 'Points'
    }
statList['Max AP Repolarization (mV)'] = {
    'name': 'postSpike_dvdt_min_val',
    'units': 'mV',
    'yStat': 'postSpike_dvdt_min_val',
    'yStatUnits': 'mV',
    'xStat': 'postSpike_dvdt_min_pnt',
    'xStatUnits': 'Points'
    }
# todo: fix this
statList['AP Duration (ms)'] = {
    'name': 'apDuration_ms',
    'units': 'ms',
    'yStat': 'apDuration_ms',
    'yStatUnits': 'ms',
    'xStat': 'thresholdPnt',
    'xStatUnits': 'Points'
    }

# new 20210211
statList['Half Width 10 (ms)'] = {
    'name': 'widths_10',
    'yStat': 'widths_10',
    'yStatUnits': 'ms',
    'xStat': '',
    'xStatUnits': ''
    }
statList['Half Width 20 (ms)'] = {
    'name': 'widths_20',
    'yStat': 'widths_20',
    'yStatUnits': 'ms',
    'xStat': '',
    'xStatUnits': ''
    }
statList['Half Width 50 (ms)'] = {
    'name': 'widths_50',
    'yStat': 'widths_50',
    'yStatUnits': 'ms',
    'xStat': '',
    'xStatUnits': ''
    }
statList['Half Width 80 (ms)'] = {
    'name': 'widths_80',
    'yStat': 'widths_80',
    'yStatUnits': 'ms',
    'xStat': '',
    'xStatUnits': ''
    }
statList['Half Width 90 (ms)'] = {
    'name': 'widths_90',
    'yStat': 'widths_90',
    'yStatUnits': 'ms',
    'xStat': '',
    'xStatUnits': ''
    }
# sa node specific
'''
statList['Region'] = {
    'yStat': 'Region',
    'yStatUnits': '',
    'xStat': '',
    'xStatUnits': ''
    }
'''
# kymograph analysis
statList['Ca++ Delay (s)'] = {
    'yStat': 'caDelay_sec',
    'yStatUnits': 's',
    'xStat': '',
    'xStatUnits': ''
    }
statList['Ca++ Width (ms)'] = {
    'yStat': 'caWidth_ms',
    'yStatUnits': 'ms',
    'xStat': '',
    'xStatUnits': ''
    }

def getHumanFromStat(statName):
    """Get human readable stat name from backend stat name."""
    ret = ''
    for k,v in statList.items():
        name = v['name']
        if name == statName:
            ret = k
            break
    if ret =='':
        logger.warning(f'Did not find "{statName}"')
    #
    return ret

def getStatFromHuman(humanStat):
    """Get backend stat name from human readable stat."""
    ret = ''
    try:
        ret = statList[humanStat]['name']
    except (KeyError) as e:
        logger.warning(f'Bad key "{humanStat}"')
    #
    return ret

def _print():
    """
    Print out human readable detection parameters and convert to markdown table

    Requires:
        pip install tabulate
    """
    import pandas as pd

    # statList is a dict with keys, one key per stat
    d = statList

    dictList = []
    for k,v in d.items():
        stat = k
        oneDict = {
            'Stat': stat,
        }
        for k2,v2 in v.items():
            oneDict[k2] = v2

        dictList.append(oneDict)
    #
    df = pd.DataFrame(dictList)
    str = df.to_markdown()
    print(str)

class bAnalysisUtil:
    def __init__(self):
        self.configDict = self.configDefault()

        # load preferences
        if getattr(sys, 'frozen', False):
            # we are running in a bundle (frozen)
            bundle_dir = sys._MEIPASS
        else:
            # we are running in a normal Python environment
            bundle_dir = os.path.dirname(os.path.abspath(__file__))
        self.configFilePath = os.path.join(bundle_dir, 'AnalysisApp_Config.json')

        self.configLoad()

        # self.top = None  # used by tkinter interface

    @staticmethod
    def getStatList():
        """20210929 working on extending stat list with sanpy.userAnalysis

        We first define stat list in this function as statList
        We then ask for user defined stat functions static members xxx
        """
        coreStatList = statList  # from global withing function

        #userStatList = sanpy.userAnalysis.findUserAnalysis()
        #userStatList = sanpy.userAnalysis.baseUserAnalysis.findUserAnalysis()
        #userStatList = findUserAnalysis()
        
        logger.warning(f'need to reactivate getting user stats')
        
        """
        Each entry in statList looks like:
        
        statList['Spike Time (s)'] = {
            'name': 'thresholdSec',
            'units': 's',
            'yStat': 'thresholdVal',
            'yStatUnits': 'mV',
            'xStat': 'thresholdSec',
            'xStatUnits': 's'
            }
        """
        userStatList : List[dict] = sanpy.user_analysis.baseUserAnalysis.findUserAnalysisStats()
        for userStatDict in userStatList:
            for k, v in userStatDict.items():
                if k in coreStatList.keys():
                    logger.error(f'user analysis key "{k}" already exists in core stat List')
                else:
                    # name = v['name']
                    # coreStatList[k] = {}
                    # coreStatList[k][name] = name
                    coreStatList[k] = v

        return coreStatList

        #20220630, was this
        userStatList = sanpy.user_analysis.baseUserAnalysis.findUserAnalysis()

        for k, v in userStatList.items():
            # check if key exists !!!
            if k in coreStatList.keys():
                logger.error(f'user analysis key {"k"} already exists in core stat List')
            else:
                name = v['name']
                coreStatList[k] = {}
                coreStatList[k][name] = name

        return coreStatList

    '''
    @staticmethod
    def humanStatToBackend(theStat):
        return ''
    '''

    def prettyPrint(self):
        print(json.dumps(self.configDict, indent=4, sort_keys=True))

    def getDetectionConfig(self):
        return self.config

    def getDetectionParam(self, theParam):
        if theParam in self.configDict['detection'].keys():
            return self.configDict['detection'][theParam]['value']
        else:
            print('error: bAnalysisUtil.getDetectionParam() detection parameter not found:', theParam)
            return None

    def getDetectionDescription(self, theParam):
        if theParam in self.configDict['detection'].keys():
            return self.configDict['detection'][theParam]['meaning']
        else:
            print('error: bAnalysisUtil.getDetectionDescription() detection parameter not found:', theParam)
            return None

    def setDetectionParam(self, theParam, theValue):
        if theParam in self.configDict['detection'].keys():
            self.configDict['detection'][theParam]['value'] = theValue
        else:
            print('error: bAnalysisUtil.setDetectionParam() detection parameter not found:', theParam)
            return None

    def configSave(self):
        print('bAnalysisUtil.configSave()')
        with open(self.configFilePath, 'w') as outfile:
            json.dump(self.configDict, outfile, indent=4, sort_keys=True)

    def configLoad(self):
        if os.path.isfile(self.configFilePath):
            print('    bAnalysisUtil.configLoad() loading configFile file:', self.configFilePath)
            with open(self.configFilePath) as f:
                self.configDict = json.load(f)
        else:
            #print('    bAnalysisUtil.preferencesLoad() using program provided default options')
            self.configDefault()

    def configDefault(self):
        theRet = OrderedDict()

        theRet['detection'] = OrderedDict()

        theRet['detection']['dvdtThreshold'] = OrderedDict()
        theRet['detection']['dvdtThreshold'] = {
            'value': 100,
            'meaning': 'Threshold crossing in dV/dt',
        }

        theRet['detection']['minSpikeVm'] = OrderedDict()
        theRet['detection']['minSpikeVm'] = {
            'value': -20,
            'meaning': 'Minimum Vm to accept a detected spike',
        }

        theRet['detection']['medianFilter'] = OrderedDict()
        theRet['detection']['medianFilter'] = {
            'value': 5,
            'meaning': 'Median filter for Vm (must be odd)',
        }

        theRet['detection']['minISI_ms'] = OrderedDict()
        theRet['detection']['minISI_ms'] = {
            'value': 75,
            'meaning': 'Minimum allowable inter-spike-interval (ms), anything shorter than this will be rejected',
        }

        return theRet

if __name__ == '__main__':
    if 0:
        _fullStatList = sanpy.bAnalysisUtil.getStatList()
        for oneStat in _fullStatList:
            sanpy._util.pprint(oneStat)
    
    if 0:
        # unit tests
        bau = bAnalysisUtil()
        print('dvdtThreshold:', bau.getDetectionParam('dvdtThreshold'))
        print('minSpikeVm:', bau.getDetectionParam('minSpikeVm'))
        print('medianFilter:', bau.getDetectionParam('medianFilter'))
        bau.configSave()

        bau.prettyPrint()

    _print()

    # abb removed 20201109, not using this, write a PyQt preferences panel
    '''
    myRoot = tkinter.Tk()
    bau.tk_PreferenesPanel(myRoot)
    myRoot.mainloop()
    '''



# 20210522

# from PyQt5 import QtCore, QtWidgets, QtGui
from PyQt5 import QtWidgets

"""
General purpose dialogs
"""

def okDialog(message, informativeText=None):
	"""
	message:
	"""
	msg = QtWidgets.QMessageBox()
	msg.setIcon(QtWidgets.QMessageBox.Warning)
	msg.setText(message)
	if informativeText is not None:
		msg.setInformativeText(informativeText)
	msg.setWindowTitle(message)
	retval = msg.exec_()
	return retval

def okCancelDialog(message, informativeText=None):
	msg = QtWidgets.QMessageBox()
	msg.setIcon(QtWidgets.QMessageBox.Warning)
	msg.setText(message)
	if informativeText is not None:
		msg.setInformativeText(informativeText)
	msg.setWindowTitle(message)
	msg.setStandardButtons(QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
	msg.setDefaultButton(QtWidgets.QMessageBox.Ok)
	retval = msg.exec_()
	return retval

def yesNoCancelDialog(message, informativeText=None):
	msg = QtWidgets.QMessageBox()
	msg.setIcon(QtWidgets.QMessageBox.Warning)
	msg.setText(message)
	if informativeText is not None:
		msg.setInformativeText(informativeText)
	msg.setWindowTitle(message)
	msg.setStandardButtons(QtWidgets.QMessageBox.Yes
							| QtWidgets.QMessageBox.No
							| QtWidgets.QMessageBox.Cancel
							)
	retval = msg.exec_()
	return retval



import math
import os
from functools import partial
import json
from typing import Union, Dict, List, Tuple, Optional, Optional

from PyQt5 import QtCore, QtWidgets, QtGui

import sanpy
from sanpy.interface.plugins import sanpyPlugin

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

class detectionParams(sanpyPlugin):
    """
    Plugin to display overview of detection parameters for analysis.

    Uses:
        QTableView: sanpy.interface.bErrorTable.errorTableView()
        QAbstractTableModel: sanpy.interface.bFileTable.pandasModel
    """
    myHumanName = 'Detection Parameters'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        logger.info('')
        
        # this is a class that holds all premade and loaded presets
        # it exists when runnin the sanpy app
        # if not running sanpy app then build it here (for debuggin)
        self._detectionObject : sanpy.bDetection = None
        
        _detectionDict : dict = None
        _currentDetectionName : str = None

        _sanPyApp = self.getSanPyApp()
        if _sanPyApp is not None:
            self._detectionObject = _sanPyApp.getDetectionClass()
        else:
            self._detectionObject = sanpy.bDetection()

        # get from ba
        if self.ba is not None:
            # get the ba detection dict
            # only exists if analysis has been performed
            _detectionDict = self.ba.getDetectionDict(asCopy=True)  # can be None

        # or, get from app
        if _detectionDict is None:
            # ba is either None or does not have detection yet
            # get the list
            presetList = self._detectionObject.getDetectionPresetList()
            # select first in list
            _currentDetectionName = presetList[0]
            _detectionDict = self._detectionObject.getDetectionDict(_currentDetectionName)        

        # deep copy because we are modifying this
        # do not want to change ba (for example) until we detect
        self._detectionDict = _detectionDict  # dict

        # the name of the detection dict in ba OR the first detection type in detection class
        _currentDetectionName = _detectionDict['detectionName']
        self._currentDetectionName = _currentDetectionName

        self._qLabelRed = 'QLabel { background-color : #888822;}'
        self._qLabelBackground = 'QLabel { background-color : #222222;}'

        self.buildUI()

        #self.insertIntoScrollArea()

    def buildUI(self):

        self.vLayout = QtWidgets.QVBoxLayout()

        # top controls
        hControlLayout = QtWidgets.QHBoxLayout()

        fileName = 'None'
        if self.ba is not None:
            fileName = self.ba.fileLoader.filename
        self.fileNameLabel = QtWidgets.QLabel(f'File: {fileName}')
        hControlLayout.addWidget(self.fileNameLabel, alignment=QtCore.Qt.AlignLeft)

        aName = 'Detect'
        aButton = QtWidgets.QPushButton(aName)
        aButton.clicked.connect(partial(self.on_button_click, aName))
        hControlLayout.addWidget(aButton, alignment=QtCore.Qt.AlignLeft)

        aName = 'Set Defaults'
        aButton = QtWidgets.QPushButton(aName)
        aButton.clicked.connect(partial(self.on_button_click, aName))
        hControlLayout.addWidget(aButton, alignment=QtCore.Qt.AlignLeft)

        # get list of names of detection presets (like SANode)
        #detectionPresetList = self.getSanPyApp().getDetectionClass().getDetectionPresetList()
        detectionPresetList = self._detectionObject.getDetectionPresetList()

        aComboBox = QtWidgets.QComboBox()
        #detectionPresets = sanpy.bDetection.getDetectionPresetList()
        for detectionPreset in detectionPresetList:
            aComboBox.addItem(detectionPreset)
        aComboBox.setCurrentText(detectionPresetList[0])
        aComboBox.currentTextChanged.connect(self.on_select_detection_preset)
        hControlLayout.addWidget(aComboBox, alignment=QtCore.Qt.AlignLeft)

        aName = 'Save Params'
        aButton = QtWidgets.QPushButton(aName)
        aButton.clicked.connect(partial(self.on_button_click, aName))
        hControlLayout.addWidget(aButton, alignment=QtCore.Qt.AlignLeft)

        hControlLayout.addStretch()

        '''
        aName = 'Set Defaults'
        aButton = QtWidgets.QPushButton(aName)
        aButton.clicked.connect(partial(self.on_button_click, aName))
        hControlLayout.addWidget(aButton)

        aName = 'SA Node'
        aButton = QtWidgets.QPushButton(aName)
        aButton.clicked.connect(partial(self.on_button_click, aName))
        hControlLayout.addWidget(aButton)

        aName = 'Ventricular'
        aButton = QtWidgets.QPushButton(aName)
        aButton.clicked.connect(partial(self.on_button_click, aName))
        hControlLayout.addWidget(aButton)

        aName = 'Ca++ Kymograph'
        aButton = QtWidgets.QPushButton(aName)
        aButton.clicked.connect(partial(self.on_button_click, aName))
        hControlLayout.addWidget(aButton)

        aName = 'Neuron'
        aButton = QtWidgets.QPushButton(aName)
        aButton.clicked.connect(partial(self.on_button_click, aName))
        hControlLayout.addWidget(aButton)

        aName = 'Subthreshold'
        aButton = QtWidgets.QPushButton(aName)
        aButton.clicked.connect(partial(self.on_button_click, aName))
        hControlLayout.addWidget(aButton)

        aName = 'Ca++ Spikes'
        aButton = QtWidgets.QPushButton(aName)
        aButton.clicked.connect(partial(self.on_button_click, aName))
        hControlLayout.addWidget(aButton)
        '''

        #
        self.vLayout.addLayout(hControlLayout)

        self.vLayoutParams = self.buildUI2()
        self.vLayout.addLayout(self.vLayoutParams)

        #
        # scroll area again
        self._tmpWidget = QtWidgets.QWidget()
        self._tmpWidget.setLayout(self.vLayout)

        self.scrollArea = QtWidgets.QScrollArea()
        self.scrollArea.setWidgetResizable(True);
        self.scrollArea.setWidget(self._tmpWidget)

        self.finalLayout = QtWidgets.QVBoxLayout()
        self.finalLayout.addWidget(self.scrollArea)

        # finalize
        #self.setLayout(self.finalLayout)
        _vBoxLayout = self.getVBoxLayout()
        _vBoxLayout.addLayout(self.finalLayout)

    def buildUI2(self):
        """Build a control for each detection parameter.
        """
        # list of keys/columns in main analysis dir file list
        analysisDirDict = sanpy.analysisDir.sanpyColumns

        dDict = self._detectionObject.getMasterDict(self._currentDetectionName)
        
        vLayoutParams = QtWidgets.QGridLayout()

        self.widgetDict = {}  # list of widgets we are displaying

        row = 0
        col = 0
        rowSpan = 1
        colSpan = 1
        for k,v in dDict.items():
            col = 0

            currentValue = self._detectionDict[k]

            # k is the name in spike dictionary
            paramName = k
            defaultValue = v['defaultValue']
            valueType = v['type']  # from ('int', 'float', 'boolean', 'string', detectionTypes_)
            units = v['units']
            humanName = v['humanName']
            description = v['description']
            allowNone = v['allowNone']  # set special value for spin box

            # flag with red
            flagWithRed = currentValue != defaultValue

            inAnalysisDir = paramName in analysisDirDict.keys()

            # add an '*' for params in main file table keys/columns
            aName = ' '
            if inAnalysisDir:
                aName = '*'
            nameLabel = QtWidgets.QLabel(aName)  # added to list so we can later set color
            col += 1

            # for debugging
            '''
            aLabel = QtWidgets.QLabel(paramName)
            vLayoutParams.addWidget(aLabel, row, col, rowSpan, colSpan)
            col += 1
            '''

            humanNameLabel = QtWidgets.QLabel(humanName)
            vLayoutParams.addWidget(nameLabel, row, col, rowSpan, colSpan)
            if flagWithRed:
                humanNameLabel.setStyleSheet(self._qLabelRed)
            vLayoutParams.addWidget(humanNameLabel, row, col, rowSpan, colSpan)
            col += 1

            # for debugging
            '''
            aLabel = QtWidgets.QLabel(valueType)
            vLayoutParams.addWidget(aLabel, row, col, rowSpan, colSpan)
            col += 1
            '''

            aLabel = QtWidgets.QLabel(units)
            vLayoutParams.addWidget(aLabel, row, col, rowSpan, colSpan)
            col += 1

            aWidget = None
            if valueType == 'int':
                aWidget = QtWidgets.QSpinBox()
                aWidget.setRange(0, 2**16)  # minimum is used for setSpecialValueText()
                aWidget.setSpecialValueText("None")  # displayed when widget is set to minimum
                if currentValue is None or math.isnan(currentValue):
                    aWidget.setValue(0)
                else:
                    aWidget.setValue(currentValue)
                aWidget.setKeyboardTracking(False) # don't trigger signal as user edits
                aWidget.valueChanged.connect(partial(self.on_spin_box, paramName))
            elif valueType == 'float':
                aWidget = QtWidgets.QDoubleSpinBox()
                aWidget.setRange(-1e9, +1e9)  # minimum is used for setSpecialValueText()
                aWidget.setSpecialValueText("None")  # displayed when widget is set to minimum

                if currentValue is None or math.isnan(currentValue):
                    aWidget.setValue(-1e9)
                else:
                    aWidget.setValue(currentValue)
                aWidget.setKeyboardTracking(False) # don't trigger signal as user edits
                aWidget.valueChanged.connect(partial(self.on_spin_box, paramName))
            elif valueType == 'list':
                # text edit a list
                pass
            elif valueType in ['bool', 'boolean']:
                # popup of True/False
                aWidget = QtWidgets.QComboBox()
                aWidget.addItem('True')
                aWidget.addItem('False')
                aWidget.setCurrentText(str(currentValue))
                aWidget.currentTextChanged.connect(partial(self.on_bool_combo_box, paramName))
            elif valueType == 'string':
                # text edit
                aWidget = QtWidgets.QLineEdit(currentValue)
                aWidget.setReadOnly(True)  # for now our 1 edit widget is not editable
                #aWidget.setKeyboardTracking(False) # don't trigger signal as user edits
                #aWidget.setValidator(QIntValidator())
                #aWidget.setMaxLength(4)
                aWidget.setAlignment(QtCore.Qt.AlignLeft)
                aWidget.editingFinished.connect(partial(self.on_text_edit, aWidget, paramName))
            elif valueType == 'sanpy.bDetection.detectionTypes':
                aWidget = QtWidgets.QComboBox()
                detectionTypes = sanpy.bDetection.detectionTypes
                for theType in detectionTypes:
                    aWidget.addItem(theType.name)
                aWidget.setCurrentText(str(currentValue))
                aWidget.currentTextChanged.connect(partial(self.on_detection_type_combo_box, paramName))
            else:
                logger.error(f'Did not understand valueType:"{valueType}" for parameter:"{paramName}"')

            if aWidget is not None:
                
                # keep track of what we are displaying
                # used to we can get default value (form 'key' and set label to red
                self.widgetDict[paramName] = {'widget': aWidget, 'nameLabelWidget':humanNameLabel}
                
                vLayoutParams.addWidget(aWidget, row, col, rowSpan, colSpan)
            #
            col += 1

            # add a 'None' button for detection parameters that allow it
            if allowNone:
                aName = f'{paramName}_none'
                aButton = QtWidgets.QPushButton('Set None')
                aButton.clicked.connect(partial(self.on_button_click, aName))
                vLayoutParams.addWidget(aButton, row, col, rowSpan, colSpan)
            #
            col += 1

            aLabel = QtWidgets.QLabel(description)
            vLayoutParams.addWidget(aLabel, row, col, rowSpan, colSpan)

            #
            row += 1
        #
        return vLayoutParams

    def _setDict(self, detectionParam, value):
        """Respond to user chhanging a detection parameter.
        
        This is getting triggered when we update values in Widgets in replot()
        Not intended but just be aware
        """
        logger.info(f'Setting dDict key:"{detectionParam}" to value "{value}" type:{type(value)}')
        if value == -1e9:
            #print(f'TWEAK paramName:{paramName} --->>> value:', value)
            value = None
        #ok = self.detectionClass.setValue(paramName, value)
        # if not ok:
        #     logger.error('')
        self._detectionDict[detectionParam] = value

        _masterDict = self._detectionObject.getMasterDict(self._currentDetectionName)
        defaultValue = _masterDict[detectionParam]['defaultValue']
        flagWithRed = value != defaultValue
        aLabel = self.widgetDict[detectionParam]['nameLabelWidget']
        if flagWithRed:
            aLabel.setStyleSheet(self._qLabelRed)
        else:
            aLabel.setStyleSheet(self._qLabelBackground)

    def on_bool_combo_box(self, paramName, text):
        #logger.info(f'paramName:{paramName} text:"{text}" {type(text)}')
        value = text == 'True'
        self._setDict(paramName, value)

    def on_detection_type_combo_box(self, paramName, text):
        #logger.info(f'paramName:"{paramName}" text:"{text}"')
        self._setDict(paramName, text)

    def on_text_edit(self, aWidget, paramName):
        text = aWidget.text()
        #logger.info(f'paramName:{paramName} text:"{text}"')
        self._setDict(paramName, text)

    def on_spin_box(self, paramName, value):
        """
        When QDoubldeSpinBox accepts None, value is -1e9
        """
        #logger.info(f'paramName:{paramName} value:"{value}" {type(value)}')
        self._setDict(paramName, value)

    def detect(self):
        logger.info('detecting spikes')
        
        if self.ba is None:
            return

        # spike detect
        self.ba.spikeDetect(self._detectionDict)

        logger.info(f'detected {self.ba.numSpikes} spikes')

        # update interface
        self.signalDetect.emit(self.ba)

    def on_select_detection_preset(self, detectionPresetName : str):
        """User selected a detection preset from combobox.
        """
        logger.info(f'detectionPresetName:"{detectionPresetName}"')
        #detectionPreset = sanpy.bDetection.detectionPresets[detectionPreset]
        #detectionPreset = sanpy.bDetection.detectionPresets(detectionPreset)
        #self.detectionClass.setToType(detectionPreset)
        
        # grab the detection from sanpy app
        self._detectionDict = self._detectionObject.getDetectionDict(detectionPresetName)
        self._currentDetectionName = self._detectionDict['detectionName']

        self.replot()

    def on_button_click(self, buttonName):

        if buttonName == 'Detect':
            # take our current dDict and ba.detectSpikes
            # need to signal main interface we did this
            self.detect()

        elif buttonName == 'Set Defaults':
            # set current dict to defaults from sanpy.bDetection
            self._detectionDict = self._detectionObject.getDetectionDict(self._currentDetectionName)
            self._currentDetectionName = self._detectionDict['detectionName']

            # refresh interface
            self.replot()
        
        elif buttonName == 'Save Params':
            self.save()

        # elif buttonName == 'SA Node':
        #     detectionPreset = sanpy.bDetection.detectionPresets.sanode
        #     self.detectionClass.setToType(detectionPreset)
        #     # refresh interface
        #     self.replot()
        # elif buttonName == 'Ventricular':
        #     detectionPreset = sanpy.bDetection.detectionPresets.ventricular
        #     self.detectionClass.setToType(detectionPreset)
        #     # refresh interface
        #     self.replot()
        # elif buttonName == 'Neuron':
        #     detectionPreset = sanpy.bDetection.detectionPresets.neuron
        #     self.detectionClass.setToType(detectionPreset)
        #     # refresh interface
        #     self.replot()
        # elif buttonName == 'Subthreshold':
        #     detectionPreset = sanpy.bDetection.detectionPresets.subthreshold
        #     self.detectionClass.setToType(detectionPreset)
        #     # refresh interface
        #     self.replot()
        # elif buttonName == 'Ca++ Spikes':
        #     detectionPreset = sanpy.bDetection.detectionPresets.caspikes
        #     self.detectionClass.setToType(detectionPreset)
        #     # refresh interface
        #     self.replot()
        # elif buttonName == 'Ca++ Kymograph':
        #     detectionPreset = sanpy.bDetection.detectionPresets.cakymograph
        #     self.detectionClass.setToType(detectionPreset)
        #     # refresh interface
        #     self.replot()

        elif buttonName.endswith('_none'):
            # set a parameter to None, will only work if 'allowNone'
            paramName = buttonName.replace('_none', '')
            self._setDict(paramName, None)
            self.replot()

        else:
            logger.warning(f'Button "{buttonName}" not understood.')

    def replot(self):
        #if self.ba is None:
        #    return

        logger.info('')

        # was this
        #detectionClass = self.detectionClass
        detectionDict = self._detectionDict

        _masterDict = self._detectionObject.getMasterDict(self._currentDetectionName)

        #for detectionParam in detectionClass.keys():
        for detectionParam in detectionDict.keys():
            if detectionParam not in self.widgetDict.keys():
                logger.warning(f'detectionParam missing "{detectionParam}" -->> not adding to display')
                continue

            #currentValue = v['currentValue']
            #currentValue = detectionClass.getValue(detectionParam)
            currentValue = detectionDict[detectionParam]

            defaultValue = _masterDict[detectionParam]['defaultValue']
            flagWithRed = defaultValue != currentValue

            if currentValue == -1e9:
                logger.error(f'XXXXX detectionParam:{detectionParam} currentValue:{currentValue} {type(currentValue)}')
                currentValue = None

            aLabel = self.widgetDict[detectionParam]['nameLabelWidget']
            if flagWithRed:
                aLabel.setStyleSheet(self._qLabelRed)
                aLabel.update()
            else:
                aLabel.setStyleSheet(self._qLabelBackground)
                aLabel.update()

            aWidget = self.widgetDict[detectionParam]['widget']
            if isinstance(aWidget, QtWidgets.QSpinBox):
                try:
                    if currentValue is None or math.isnan(currentValue):
                        aWidget.setValue(0)
                    else:
                        aWidget.setValue(currentValue)
                except (TypeError) as e:
                    logger.error(f'QSpinBox detectionParam:{detectionParam} ... {e}')
            elif isinstance(aWidget, QtWidgets.QDoubleSpinBox):
                try:
                    if currentValue is None or math.isnan(currentValue):
                        aWidget.setValue(-1e9)
                    else:
                        aWidget.setValue(currentValue)
                except (TypeError) as e:
                    logger.error(f'QDoubleSpinBox detectionParam:{detectionParam} ... {e}')
            elif isinstance(aWidget, QtWidgets.QComboBox):
                aWidget.setCurrentText(str(currentValue))
            elif isinstance(aWidget, QtWidgets.QLineEdit):
                aWidget.setText(str(currentValue))
            else:
                logger.warning(f'key "{detectionParam}" has value "{currentValue}" but widget type "{type(aWidget)}" not understood.')

    def slot_switchFile(self,
                        ba : sanpy.bAnalysis,
                        rowDict : Optional[dict] = None,
                        replot : bool = True):
        # don't replot until we set our detectionClass
        replot = False
        super().slot_switchFile(ba, rowDict, replot=replot)
        
        # before we detect a ba, its detection dict is None
        _detectionDict = self.ba.getDetectionDict(asCopy=True)  # can be None
        if _detectionDict is not None:
            self._detectionDict = _detectionDict  # 
            self._currentDetectionName = _detectionDict['detectionName']

        self.replot()

    def save(self):
        """Save our current detection dict to a user file.
        """
        userDetectionFolder = sanpy._util._getUserDetectionFolder()

        # key 'detectionName'
        filename = self._detectionDict['userSaveName']
        if not filename:
            filename = self._detectionDict['detectionName'] + ' User'
            
        savePath = os.path.join(userDetectionFolder, filename+'.json')

        userSaveFile, _tmp = QtWidgets.QFileDialog.getSaveFileName(self, 'Save Detection Params', savePath)
        logger.info(userSaveFile)

        with open(userSaveFile, 'w') as f:
            self._detectionDict['userSaveName'] = filename
            json.dump(self._detectionDict, f, indent=4)

if __name__ == '__main__':

    path = 'data/19114001.abf'
    ba = sanpy.bAnalysis(path)

    bd = sanpy.bDetection()  # gets default
    dDict = bd.getDetectionDict('SA Node')  # an actual dict

    # from pprint import pprint
    # pprint(dDict)
    
    #dDict['dvdtThreshold'] = 12

    ba.spikeDetect(dDict)
    logger.info(f'detected numSpike:{ba.numSpikes}')

    import sys
    app = QtWidgets.QApplication([])

    dp = detectionParams(ba=ba)  # dec 2022, do not pass bd
    dp.show()

    '''
    scrollArea = dp.insertIntoScrollArea()
    if scrollArea is not None:
        scrollArea.show()
    else:
        dp.show()
    '''

    sys.exit(app.exec_())



import numpy as np

import pyabf

from sanpy.fileloaders.fileLoader_base import fileLoader_base, recordingModes

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

class fileLoader_atf(fileLoader_base):
    loadFileType = 'atf'
    
    def __init__(self, path):
        super().__init__(path)
        self._loadAtf()

    def _loadAtf(self, loadData):
        # We cant't get dataPointsPerMs without loading the data
        loadData = True

        self._abf = pyabf.ATF(self.filepath)  # , loadData=loadData)
        print('  self._abf:', self._abf.sweepList)
        #try:
        if 1:
            self._sweepList = self._abf.sweepList
            self._sweepLengthSec = self._abf.sweepLengthSec
            if loadData:
                tmpRows = self._abf.sweepX.shape[0]
                numSweeps = len(self._sweepList)
                self._sweepX = np.zeros((tmpRows,numSweeps))
                self._sweepY = np.zeros((tmpRows,numSweeps))
                self._sweepC = np.zeros((tmpRows,numSweeps))

                for sweep in self._sweepList:
                    self._abf.setSweep(sweep)
                    self._sweepX[:, sweep] = self._abf.sweepX  # <class 'numpy.ndarray'>, (60000,)
                    self._sweepY[:, sweep] = self._abf.sweepY
                    #self._sweepC[:, sweep] = self._abf.sweepC
                # not needed
                self._abf.setSweep(0)

            logger.info(f'Loaded abf self._sweepX: {self._sweepX.shape}')
            # get v from pyAbf
            if len(self.sweepX.shape) > 1:
                t0 = self.sweepX[0][0]
                t1 = self.sweepX[1][0]
            else:
                t0 = self.sweepX[0]
                t1 = self.sweepX[1]
            dt = (t1-t0) * 1000
            dataPointsPerMs = 1 / dt
            dataPointsPerMs = round(dataPointsPerMs)
            self._dataPointsPerMs = dataPointsPerMs

            '''
            channel = 0
            adcUnits = self._abf.adcUnits[channel]
            self._sweepLabelY = adcUnits
            self._sweepLabelX = "sec'"
            '''
            self._sweepLabelY = 'mV'
            self._sweepLabelX = "sec'"

            if self._sweepLabelY in ['pA']:
                self._recordingMode = recordingModes.vclamp  # 'V-Clamp'
            elif self._sweepLabelY in ['mV']:
                self._recordingMode = recordingModes.iclamp  #'I-Clamp'

        '''
        except (Exception) as e:
            # some abf files throw: 'unpack requires a buffer of 234 bytes'
            logger.error(f'did not load ATF file: {self._path}')
            logger.error(f'  unknown Exception was: {e}')
            self.loadError = True
            self._abf = None
        '''

        self.myFileType = 'atf'

        # don't keep _abf, we grabbed every thing we needed



import numpy as np
import scipy.signal

from PyQt5 import QtCore, QtGui, QtWidgets

import matplotlib.pyplot as plt

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

import sanpy
from sanpy.interface.plugins import sanpyPlugin

from sanpy.bAnalysisUtil import statList

class basePlotTool(sanpyPlugin):
    """
    """
    myHumanName = 'Base Plot Tool'

    def __init__(self, **kwargs):
        super(basePlotTool, self).__init__(**kwargs)

        self.masterDf = None
        # one
        #self.masterDf = self.ba.dfReportForScatter
        # pool of bAnalysis
        #self.masterDf = self._bPlugins._sanpyApp.myAnalysisDir.pool_build()

        self.mainWidget2 = None

        # turn off all signal/slot
        switchFile = self.responseTypes.switchFile
        self.toggleResponseOptions(switchFile, newValue=False)
        analysisChange = self.responseTypes.analysisChange
        self.toggleResponseOptions(analysisChange, newValue=False)
        selectSpike = self.responseTypes.selectSpike
        self.toggleResponseOptions(selectSpike, newValue=False)
        setAxis = self.responseTypes.setAxis
        self.toggleResponseOptions(setAxis, newValue=False)

        #self.plot()

    def plot(self):
        logger.info('')

        # this plugin does not rely on underlying self.ba
        #if self.ba is None:
        #    return

        # analysisName = 'analysisname'
        analysisName = 'file'
        statListDict = statList # maps human readable to comments
        categoricalList = ['include', 'condition', 'cellType', 'sex', 'file', 'File Number']#, 'File Name']
        hueTypes = ['cellType', 'sex', 'condition', 'file', 'File Number'] #, 'File Name'] #, 'None']
        sortOrder = ['cellType', 'sex', 'condition', 'File Number']
        
        limitToCol = ['epoch']
        
        interfaceDefaults = {'Y Statistic': 'Spike Frequency (Hz)',
                            'X Statistic': 'Spike Frequency (Hz)',
                            'Hue': 'cellType',
                            'Group By': 'file'}

        #analysisName, masterDf = analysisName, df0 = ba.getReportDf(theMin, theMax, savefile)

        #print('!!!! FIX THIS in plugin plotTool.plot()')
        #masterDf = self._bPlugins._sanpyApp.myAnalysisDir.pool_build()
        # was this
        #masterDf = self.ba.dfReportForScatter

        if self.masterDf is None:
            logger.error('Did not get analysis df, be sure to run detectioon')
            return
        #bScatterPlotMainWindow
        #self.scatterWindow = sanpy.scatterwidget.bScatterPlotMainWindow(
        path = ''
        self.mainWidget2 = sanpy.interface.bScatterPlotMainWindow(
                        path, categoricalList, hueTypes,
                        analysisName, sortOrder, statListDict=statListDict,
                        masterDf = self.masterDf,
                        limitToCol = limitToCol,
                        interfaceDefaults = interfaceDefaults)
        # rewire existing widget into plugin architecture
        #self.mainWidget.closeEvent = self.onClose
        self._mySetWindowTitle()

        #self.mainWidget2.show()

        tmpLayout = QtWidgets.QVBoxLayout()
        tmpLayout.addWidget(self.mainWidget2)

        #
        #self.setLayout(tmpLayout)
        self.getVBoxLayout().addLayout(tmpLayout)
        


from functools import partial

import numpy as np
import pandas as pd

from PyQt5 import QtWidgets, QtCore

import sanpy
import sanpy.interface
import sanpy.bDetection
from sanpy.interface.bFileTable import pandasModel

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

class fileListWidget(QtWidgets.QWidget):
    """A widget showing
        - file list controls
        - file list widget, see: sanpy.interface.bTableView
    """

    signalLoadFolder = QtCore.pyqtSignal(object, object)  # path, depth

    #signalSelectRow = QtCore.pyqtSignal(object, object, object)  # (row, rowDict, selectingAgain)
    #signalUpdateStatus = QtCore.pyqtSignal(object)
    """Update status in main SanPy app."""
    #signalSetDefaultDetection = QtCore.pyqtSignal(object, object)  # selected row, detection type

    def __init__(self, myModel : pandasModel, parent=None):
        """
        Args:
            myModel: sanpy.interface.bFileTable.pandasModel
        """
        super().__init__(parent)
        self._myModel : pandasModel = myModel
        # self.setAcceptDrops(True)
        self._folderDepth = 1
        self._buildUI()

    # def dragEnterEvent(self, event):
    #     logger.info('')
    
    # def dropEvent(self, event):
    #     logger.info('')

    def mySetModel(self, theModel):
        """
        Args:
            theModel: sanpy.interface.bFileTable.pandasModel
        """
        self._myModel = theModel
        self._tableView.mySetModel(theModel)

    def getTableView(self):
        return self._tableView

    def _buildUI(self):
        self._vLayout = QtWidgets.QVBoxLayout()

        self._hToolbarLayout = QtWidgets.QHBoxLayout()

        buttonName = 'Load Folder'
        button = QtWidgets.QPushButton(buttonName)
        #button.setToolTip('Save Detected Spikes to Excel file')
        button.clicked.connect(partial(self.on_button_click,buttonName))
        self._hToolbarLayout.addWidget(button, alignment=QtCore.Qt.AlignLeft)

        labelName = 'Depth'
        aLabel = QtWidgets.QLabel(labelName)
        self._hToolbarLayout.addWidget(aLabel)

        aSpinBox = QtWidgets.QSpinBox()
        aSpinBox.setMinimum(1)
        aSpinBox.setMaximum(5)
        aSpinBox.setValue(1)
        aSpinBox.valueChanged.connect(self.on_depth_spin_box)
        self._hToolbarLayout.addWidget(aSpinBox, alignment=QtCore.Qt.AlignLeft)

        self._hToolbarLayout.addStretch()

        self._vLayout.addLayout(self._hToolbarLayout)

        #
        # actual list of file
        self._tableView = sanpy.interface.bTableView(self._myModel)
        self._vLayout.addWidget(self._tableView)

        self.setLayout(self._vLayout)

    def on_depth_spin_box(self, value):
        self._folderDepth = value

    def on_button_click(self, name):
        logger.info(f'{name}')
        if name == 'Load Folder':
            self.signalLoadFolder.emit(None, self._folderDepth)

    def getDepth(self):
        """Get the folder depth.
        """
        return self._folderDepth







import os

import pandas as pd
import numpy as np

import sanpy

import sanpy.interface.plugins.stimGen2
from sanpy.interface.plugins.stimGen2 import readFileParams, buildStimDict

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

"""
bAnalysis class that loads a stim file.

TODO:
	- as usual, split into interface and gui.
		For example, move this to main api (move out of interface)
		sanpy.interface.plugins.stimGen2.readCommentParams(stimFileComment)
"""

class bAnalysisStim(sanpy.bAnalysis):
	def __init__(self, path, loadData=True, stimulusFileFolder=None):
		if stimulusFileFolder is None:
			stimulusFileFolder = os.path.split(path)[0] #'/media/cudmore/data/stoch-res' #os.path.split(path)[0]
		super(bAnalysisStim, self).__init__(file=path, loadData=True, stimulusFileFolder=stimulusFileFolder)

		print(self)
		print(self.abf)  # will be none
		#tmpSweepC = self.abf.sweepC

		#stimFilePath, stimFileComment = pyabf.stimulus.abbCommentFromFile(self.abf)
		stimFilePath, stimFileComment = self._commentFromAtfFile()
		self._stimDict = None  # raw parameters from snapy.interface.plugins.stimGen
		self._stimFile = None
		self._stimDictList = None  # one element per sweep
		self._stimFileDf = None
		if stimFileComment is not None:
			self._stimFile = stimFilePath
			self._stimDict = sanpy.interface.plugins.stimGen2.readCommentParams(stimFileComment)
			self._stimDict['file'] = os.path.split(stimFilePath)[1]  # atf does not know abf file name
			# TODO: there is an error in saving atf comment string, post sweeps are always 0
			# print(f'self._stimDict:', self._stimDict)
			self._stimDictList = sanpy.interface.plugins.stimGen2.buildStimDict(self._stimDict, path=self.path)
			self._stimFileDf = self._stimFileAsDataFrame()

	def isiStats(self, hue='sweep'):
		"""
		Get mean/sd/sem etc after grouping by 'sweep'
		"""
		if self.numSpikes == 0:
			return None

		#df = self.analysisDf
		df = self.asDataFrame()
		#print('df:', df)
		if df is None:
			return pd.DataFrame()  # empty

		df = self.reduce()  # reduce based on start/stop of sin stimulus

		statList = []

		statStr = 'isi_ms'

		stimFileDf = self._stimFileDf

		hueList = df[hue].unique()
		#for idx,oneHue in enumerate(hueList):
		for idx,sweep in enumerate(range(self.numSweeps)):
			oneHue = sweep

			dfPlot = df[ df[hue]==oneHue ]  # will be empty if 0 spikes in sweep

			#someSpikes = not dfPlot.empty # no spikes for this sweep

			ipi_ms = dfPlot[statStr]

			numIntervals = len(ipi_ms)
			#print('numIntervals:', numIntervals)

			minISI = np.min(ipi_ms) if numIntervals>1 else float('nan')
			maxISI = np.max(ipi_ms) if numIntervals>1 else float('nan')

			meanISI = np.nanmean(ipi_ms) if numIntervals>1 else float('nan')
			medianISI = np.nanmedian(ipi_ms) if numIntervals>1 else float('nan')
			stdISI = np.nanstd(ipi_ms) if numIntervals>1 else float('nan')
			cvISI = stdISI / meanISI if numIntervals>1 else float('nan')
			cvISI = round(cvISI, 3) if numIntervals>1 else float('nan')
			cvISI_inv = np.nanstd(1/ipi_ms) / np.nanmean(1/ipi_ms) if numIntervals>1 else float('nan')
			cvISI_inv = round(cvISI_inv, 3) if numIntervals>1 else float('nan')

			oneDict = {
				'file': self.getFileName(),
				'stat': statStr,
				'sweep': idx,
				'count': numIntervals,
				'minISI': minISI, #round(np.nanmin(ipi_ms),3),
				'maxISI': maxISI, #round(np.nanmax(ipi_ms),3),
				'stdISI': round(stdISI,3),
				'meanISI': round(meanISI,3),
				'medianISI': round(medianISI,3),
				'cvISI': cvISI,
				'cvISI_inv': cvISI_inv,

				# stim params from stimulus file
				'Stim Freq (Hz)': '',
				'Stim Amp': '',
				'Noise Amp': '',

			}

			if stimFileDf is not None:
				if idx < len(stimFileDf):
					oneDict['Stim Freq (Hz)'] = stimFileDf.loc[idx, 'freq(Hz)']
					oneDict['Stim Amp'] = stimFileDf.loc[idx, 'amp']
					oneDict['Noise Amp'] = stimFileDf.loc[idx, 'noise amp']
				else:
					# This happens when pClamp sweeps > stimGen sweeps (like forgetting to set post=1 in stimgen)
					logger.error(f'Did not find row {idx} in stimFileDict')

			statList.append(oneDict)

		#
		retDf = pd.DataFrame(statList)
		return retDf

	def _commentFromAtfFile(self, channel=0):
		"""
		Get the comment string from atf stimulus file
		"""
		from pyabf.stimulus import findStimulusWaveformFile
		from pyabf.stimulus import cachedStimuli

		stimPath = findStimulusWaveformFile(self.abf, channel)  # , verbose=False)
		if stimPath is None:
			return None, None
		else:
			return stimPath, cachedStimuli[stimPath].header['comment']

	@property
	def stimDict(self):
		"""
		If recorded with a stimulus file, return dict of stimulus parameters.
		Stimulus parameters are defined in sanpy/interface/plugins/stimGen2.py
		"""
		return self._stimDict

	def reduce(self):
		"""
		Reduce spikes based on start/stop of sin stimulus

		TODO:
			Use readFileParams(atfPath) to read stimulus ATF file and get start/stop

		Return:
			df of spikes within sin stimulus
		"""
		#print(self.stimDict)

		#df = self.analysisDf
		df = self.asDataFrame()  # if no spikes, will be empty

		startSec, stopSec = self.getStimStartStop()

		if startSec is not None and stopSec is not None:
			df = df[ (df['thresholdSec']>=startSec) & (df['thresholdSec']<=stopSec)]

		'''
		if self.stimDict is not None:
			#print(ba.stimDict)
			startSec = self.stimDict['stimStart_sec']
			durSeconds = self.stimDict['stimDur_sec']
			stopSec = startSec + durSeconds
			try:
				df = df[ (df['peak_sec']>=startSec) & (df['peak_sec']<=stopSec)]
			except (KeyError) as e:
				logger.error(e)
		'''
		#
		return df

	def getStimStartStop(self):
		"""
		If recorded with a stimGen atf file, get start/stop of stimulus.
		Some sweeps will be black, search for the first sweep that has a stimulus parameters
		assuming the start/stop of each sweep is identical (may not be in the future)

		Returns:
			startSec (float): Start sec of the stimulus
			stopSec (float): Stop sec of the stimulus
		"""
		#dfStimFile = self._stimFileAsDataFrame()  # get underlying atf file params (from stimGen2)
		dfStimFile = self._stimFileDf

		if dfStimFile is None:
			# no stim file
			return None, None

		startSec = None
		durSec = None
		stopSec = None
		for idx, sweep in enumerate(range(self.numSweeps)):
			tmpStartSec = dfStimFile.loc[sweep, 'start(s)']
			if not isinstance(tmpStartSec, str):
				startSec = tmpStartSec
				durSec = dfStimFile.loc[sweep, 'dur(s)']
				stopSec = startSec + durSec
				break
		return startSec, stopSec

	def _stimFileAsDataFrame(self):
		"""
		If recorded using stimGen atf file.
		Return the stmi parameters as a dataframe
		"""
		d = self.stimDict
		if d is None:
			return

		#dList = buildStimDict(d, path=self.filePath)
		#for one in dList:
		#	print(one)
		df = pd.DataFrame(self._stimDictList)
		return df



# 20210610

import os
import sys
#import pathlib
import glob
import importlib
import inspect

from typing import Union, Dict, List, Tuple, Optional

import sanpy
#import sanpy.interface
import sanpy.interface.plugins

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

class bPlugins():
    """Generate a dict of plugins.

    Populate the dict by looking in
        - package, sanpy.interface.plugins
        - folder, <user>/sanpy_plugins
    """
    def __init__(self, sanpyApp : Optional["sanpy.interface.SanPyWindow"] = None):
        
        self._sanpyApp = sanpyApp

        self.userPluginFolder : str = sanpy._util._getUserPluginFolder()
        if not os.path.isdir(self.userPluginFolder):
            self.userPluginFolder = None
        """path to <user>/plugin_dir"""

        self.pluginDict = {}
        """dict of pointers to open plugins"""

        self._openSet = set()
        """set of open plugins"""

        self.loadPlugins()

    def getSanPyApp(self) -> Optional["sanpy.interface.SanPyWindow"]:
        """Get the underlying SanPy app.
        """
        return self._sanpyApp

    def loadPlugins(self):
        """Load plugins from both:
         - Package: sanpy.interface.plugins
         - Folder: <user>/sanpy_plugins
        
        See: sanpy.fileLoaders.fileLoader_base.getFileLoader()
        """
        self.pluginDict = {}

        # Enum is to ignore bPlugins.py class ResponseType(Enum)
        ignoreModuleList = ['sanpyPlugin', 'myWidget',
                            'ResponseType', 'SpikeSelectEvent',
                            'basePlotTool',
                            'NavigationToolbar2QT',
                            'myStatListWidget']

        #
        # system plugins from sanpy.interface.plugins
        #print('loadPlugins sanpy.interface.plugins:', sanpy.interface.plugins)
        loadedList = []
        for moduleName, obj in inspect.getmembers(sanpy.interface.plugins):
            #print('moduleName:', moduleName, 'obj:', obj)
            if inspect.isclass(obj):
                #logger.info(f'moduleName: {moduleName}')
                if moduleName in ignoreModuleList:
                    # our base plugin class
                    continue
                loadedList.append(moduleName)
                fullModuleName = 'sanpy.interface.plugins.' + moduleName
                humanName = obj.myHumanName  # myHumanName is a static str
                showInMenu = obj.showInMenu  # showInMenu is a static bool
                pluginDict = {
                    'pluginClass': moduleName,
                    'type': 'system',
                    'module': fullModuleName,
                    'path': '',
                    'constructor': obj,
                    'humanName': humanName,
                    'showInMenu': showInMenu,
                }
                if humanName in self.pluginDict.keys():
                    logger.warning(f'Plugin already added "{moduleName}" humanName:"{humanName}"')
                else:
                    self.pluginDict[humanName] = pluginDict

        # print the loaded plugins
        logger.info(f'Loaded plugins:')
        for loaded in loadedList:
            logger.info(f'    {loaded}')
        # sort
        self.pluginDict = dict(sorted(self.pluginDict.items()))

        #
        # user plugins from files in folder <user>/SanPy/plugins
        loadedModuleList = []
        if self.userPluginFolder is not None:
            files = glob.glob(os.path.join(self.userPluginFolder, '*.py'))
        else:
            files = []
        
        for file in files:
            if file.endswith('__init__.py'):
                continue

            moduleName = os.path.split(file)[1]
            moduleName = os.path.splitext(moduleName)[0]
            fullModuleName = 'sanpy.interface.plugins.' + moduleName

            loadedModule = self._module_from_file(fullModuleName, file)

            try:
                oneConstructor = getattr(loadedModule, moduleName)
            except (AttributeError) as e:
                logger.error(f'Did not load user plugin, make sure file name and class name are the same:"{moduleName}"')
            else:
                humanName = oneConstructor.myHumanName
                showInMenu = oneConstructor.showInMenu
                pluginDict = {
                    'pluginClass': moduleName,
                    'type': 'user',
                    'module': fullModuleName,
                    'path': file,
                    'constructor': oneConstructor,
                    'humanName': humanName,
                    'showInMenu': showInMenu,
                    }
                if humanName in self.pluginDict.keys():
                    logger.warning(f'Plugin already added "{moduleName}" humanName:"{humanName}"')
                else:
                    self.pluginDict[humanName] = pluginDict
                    loadedModuleList.append(moduleName)

        # print the loaded plugins
        logger.info(f'Loaded {len(loadedModuleList)} user plugin(s)')
        for loaded in loadedModuleList:
            logger.info(f'    {loaded}')

    def runPlugin(self,
                  pluginName : str,
                  ba : sanpy.bAnalysis,
                  show : bool = True):
        """Run one plugin with given a bAnalysis.

        Args:
            pluginName (str):
            ba (bAnalysis): object
            show:
        """
        if not pluginName in self.pluginDict.keys():
            logger.error(f'Did not find plugin: "{pluginName}"')
            return
        else:
            humanName = self.pluginDict[pluginName]['constructor'].myHumanName

            ltwhTuple = None
            
            # get the visible x-axis from main app
            startStop = None
            app = self.getSanPyApp()
            _pluginDict = None
            if app is not None:
                startSec = app.startSec
                stopSec = app.stopSec
                if startSec is None and stopSec is None:
                    startStop= None
                else:
                    startStop = [startSec, stopSec]

                # has keys for (l, t, w, h)
                _pluginDict = app.getOptions().preferencesGet('pluginPanels', humanName)
                # ltwhTuple = None
                # if _pluginDict is not None:
                #     ltwhTuple = tuple(
                #         _pluginDict['l'],
                #         _pluginDict['t'],
                #         _pluginDict['w'],
                #         _pluginDict['h'],
                #     )
            logger.info(f'Running plugin:')
            logger.info(f'  pluginName:{pluginName}')
            logger.info(f'  humanName:{humanName}')
            logger.info(f'  startStop:{startStop}')
            if _pluginDict is not None:
                logger.info(f'  _pluginDict:{_pluginDict}')
            logger.info(f'  TODO: put try except back in !!!')
            # TODO: to open PyQt windows, we need to keep a local (persistent) variable
            #try:
            if 1:
                #print(1)
                newPlugin = \
                    self.pluginDict[pluginName]['constructor'](ba=ba, bPlugin=self, startStop=startStop)
                #print(2)
                if not newPlugin.getInitError() and show:
                    if _pluginDict is not None:
                        newPlugin.setGeometry(_pluginDict['l'],
                                            _pluginDict['t'],
                                            _pluginDict['w'],
                                            _pluginDict['h'])
                    newPlugin.getWidget().show()
                    # newPlugin.getWidget().raise_()  # bring to front, raise is a python keyword
                    # newPlugin.getWidget().activateWindow()  # bring to front

                # add the plugin to open next time we run
                
            # except(TypeError) as e:
            #     logger.error(f'Error opening plugin "{pluginName}": {e}')
            #     return
            if not newPlugin.getInitError():
                self._openSet.add(newPlugin)

            return newPlugin

    def getType(self, pluginName):
        """
        returns one of ('system', 'user')
        """
        if not pluginName in self.pluginDict.keys():
            logger.error(f'Did not find plugin: "{pluginName}"')
            return None
        else:
            theRet = self.pluginDict[pluginName]['type']
            return theRet

    def pluginList(self, forMenu=True):
        """Get list of names of loaded plugins

        Pay attention to showInMenu
        """
        retList = []
        for k,v in self.pluginDict.items():
            # k is the name of the plugin
            # v is a dict with its current state
            # print(' QANON ', k)
            # print('     v:', v)
            if not v['showInMenu']:
                continue
            retList.append(k)
        return retList

    def getHumanNames(self):
        retList = []
        for k,v in self.pluginDict.items():
            myHumanName = v['myHumanName']
            retList.append(myHumanName)
        return retList

    def _module_from_file(self, module_name : str, file_path : str):
        """
        
        Args:
            module_name: Is like sanpy.interface.plugins.onePluginFile
            file_path: Full path to onePluginFile source code (onePluginFile.py)
        """
        spec = importlib.util.spec_from_file_location(module_name, file_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module

    def slot_closeWindow(self, pluginObj):
        """Close named plugin window.

        Args:
            pluginObj (object): The running plugin object reference.

        Important:
            Need to disconnect signal/slot using _disconnectSignalSlot().
            Mostly connections to main SanPy app signals
        """
        #logger.info(pluginObj)
        try:
            logger.info(f'Removing plugin from _openSet: "{pluginObj.getHumanName()}"')
            # Critical to detatch signal/slot, removing from set does not seem to do this?
            pluginObj._disconnectSignalSlot()
            self._openSet.remove(pluginObj)

            # remove from preferences
            if pluginObj is not None and self.getSanPyApp() is not None:
                self.getSanPyApp().configDict.removePlugin(pluginObj.getHumanName())

        except (KeyError) as e:
            logger.exception(e)

    def slot_selectSpike(self, sDict):
        """
        On user selection of spike in a plugin

        Tell main app to select a spike everywhere
        """
        logger.info(sDict)
        app = self.getSanPyApp()
        if app is not None:
            spikeNumber = sDict['spikeNumber']
            doZoom = sDict['doZoom']
            app.selectSpike(spikeNumber, doZoom=doZoom)

    def slot_selectSpikeList(self, sDict : dict):
        """On user selection of spike(s) in a plugin

        Tell main app to select a spike everywhere

        Args:
            sDict (dict) with keys
                spikeList : List[int]
                doZoom : bool
                ba : sanpy.bAnalysis_.bAnalysis
        """
        logger.info(sDict)
        app = self.getSanPyApp()
        if app is not None:
            spikeList = sDict['spikeList']
            doZoom = sDict['doZoom']
            app.selectSpikeList(spikeList, doZoom=doZoom)

def test_print_classes():
    """testing"""
    print('__name__:', __name__)
    for name, obj in inspect.getmembers(sys.modules[__name__]):
        if inspect.isclass(obj):
            print(name, ':', obj)

    print('===')
    for name, obj in inspect.getmembers(sanpy.interface.plugins):
        if inspect.isclass(obj):
            print(name, ':', obj)

if __name__ == '__main__':
    #print_classes()
    #sys.exit(1)

    bp = bPlugins()

    pluginList = bp.pluginList()
    print('pluginList:', pluginList)

    abfPath = '/Users/cudmore/Sites/SanPy/data/19114001.abf'
    ba = sanpy.bAnalysis(abfPath)

    bp.runPlugin('plotRecording', ba)
    #bp.runPlugin('plotRecording3', ba)



"""
Widget to act as the interface for kymographPlugin

I am keeping it seperate because it can be used standalone.
"""

import os
import sys
from turtle import back
import numpy as np

from qtpy import QtGui, QtSql, QtCore, QtWidgets

import pyqtgraph as pg

#import qdarkstyle

# turn off qdarkstyle logging
#import logging
#logging.getLogger('qdarkstyle').setLevel(logging.WARNING)

#from shapeanalysisplugin._my_logger import logger
#from shapeanalysisplugin import kymographAnalysis

import sanpy
import sanpy.interface
#from sanpy import kymographAnalysis

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

class kymographPlugin2(QtWidgets.QWidget):

    def __init__(self, ba : sanpy.bAnalysis, parent = None):
        logger.info('')
        
        super().__init__(parent)
        
        # switching to hold actual bAnalysis
        #self._ba = sanpy.bAnalysis(path)
        self._ba = None
        if ba is not None and ba.fileLoader.isKymograph():
            self._ba = ba

        # self._kymographAnalysis = sanpy.kymographAnalysis(path)
        # dec 2022
        self._kymographAnalysis = None  # will be created in slotSwitchFile()
        if self._ba is not None:
            self._kymographAnalysis = sanpy.kymAnalysis(ba.getFilePath())
        

        self._imageMedianKernel = 0
        self._lineMedianKernel = 5

        self._currentLineNumber = 0

        self._fitIsVivible = True
        # keep track of left/right fit on top of kym image

        self._kymWidgetMain = None  #sanpy.interface.kymographWidget(self._ba)
        self._initGui()

        self._on_slider_changed(0)

        self.refreshSumLinePlot()
        self.refreshDiameterPlot()

    def _checkboxCallback(self, state : bool, name : str):
        logger.info(f'state:{state} name:{name}')
        if name == 'Line Profile':
            self.showLineProfile(state)
        elif name == 'Diameter':
            self.diameterPlotItem.setVisible(state)
        elif name == 'Sum Line':
            self.sumIntensityPlotItem.setVisible(state)
        elif name == 'Fit On Image':
            self._fitIsVivible = state
            self.refreshDiameterPlot()

    def showLineProfile(self, state : bool):
        """Toggle interface for "Line Profile"
        """
        #self.profileSlider.setVisible(state)
        
        self.lineIntensityPlotItem.setVisible(state)

        for line in self._sliceLinesList:
            line.setVisible(state)

    def _buttonCallback(self, name):
        logger.info(f'name:{name}')
        # pos = self._rectRoi.pos()  # (left, bottom)
        # size = self._rectRoi.size()  # (widht, height)
        #print(f'    pos:{pos} size:{size}')
    
        if name == 'Reset ROI':
            pos, size = self._kymographAnalysis.getFullRectRoi()  # (w,h)
            self._rectRoi.setPos(pos)
            self._rectRoi.setSize(size)
            # set the x-axis of the image plot
            #self.kymographWindow.setXRange(0, shape[0])

        elif name == 'Analyze':
            self._analyzeButton.setStyleSheet("background-color : #AA2222")
            self._analyzeButton.repaint()
            self._analyzeButton.update()
            QtCore.QCoreApplication.processEvents()

            #left_idx, right_idx = self._kymographAnalysis.getLineProfileWidth()
            imageMedianKernel = self._imageMedianKernel
            lineMedianKernel = self._lineMedianKernel
            self._kymographAnalysis.analyze(imageMedianKernel=imageMedianKernel, lineMedianKernel=lineMedianKernel)
            
            #self.refreshSumLinePlot()
            self.refreshDiameterPlot()
            
            self._analyzeButton.setStyleSheet("background-color : #22AA22")
            self._analyzeButton.repaint()
            self._analyzeButton.update()
            QtCore.QCoreApplication.processEvents()

        elif name == 'Save':
            saveFilePath = self._kymographAnalysis.getAnalysisFile()
            # get file name to save
            #name = QtGui.QFileDialog.getSaveFileName(self, 'Save File')
            savefile, tmp = QtWidgets.QFileDialog.getSaveFileName(self, 'Save File', saveFilePath)
            self._kymographAnalysis.save(savefile)

        elif name == 'Rot -90':
            newImage = self._kymographAnalysis.rotateImage()
            self.kymographImage.setImage(newImage)

        else:
            logger.warning(f'did not understand "{name}"')

    def setLineWidth(self, value : int):
        logger.info(value)
        if self._kymographAnalysis is not None:
            self._kymographAnalysis.setLineWidth(value)

    def setPercentMax(self, value : float):
        logger.info(f'value:{value}')
        if self._kymographAnalysis is not None:
            self._kymographAnalysis.setPercentOfMax(value)

    def slotSwitchFile(self, ba :sanpy.bAnalysis):
        if ba is not None and not ba.fileLoader.isKymograph():
            return

        self._ba = ba
        
        self._kymWidgetMain.slot_switchFile(ba=self._ba)

        path = self._ba.getFilePath()
        self._kymographAnalysis = sanpy.kymAnalysis(path)

        self.refreshSumLinePlot()
        self.refreshDiameterPlot()

    def _slot_roi_changed2(self, newRect):
        """User modified ROI.
        
        Sent from bKymographWidget. Rewrite of '_slot_roi_changed'
            To use new improved bKymographWidget
        
        Args:
            newRect: [l,t,r,b]
        """
        lNew = newRect[0]
        tNew = newRect[1]
        rNew = newRect[2]
        bNew = newRect[3]

        widthNew = rNew - lNew
        heightNew = tNew - bNew  # swapped

        _pos = (lNew, bNew)
        _size = (widthNew, heightNew)

        # pos = self._rectRoi.pos()
        # size = self._rectRoi.size()

        logger.info(f'pos:{_pos} size:{_size}')

        self._kymographAnalysis.setPosRoi(_pos)
        self._kymographAnalysis.setSizeRoi(_size)

        # update line profile plot
        #_lineScanValue = self.profileSlider.value()
        #self._on_slider_changed(_lineScanValue)

        # replot sum
        self.refreshSumLinePlot()

    def old_slot_roi_changed(self):
        pos = self._rectRoi.pos()
        size = self._rectRoi.size()

        logger.info(f'pos:{pos} size:{size}')

        self._kymographAnalysis.setPosRoi(pos)
        self._kymographAnalysis.setSizeRoi(size)

        # update line profile plot
        _lineScanValue = self.profileSlider.value()
        self._on_slider_changed(_lineScanValue)

    def setImageMedianKernel(self, value):
        self._imageMedianKernel = value
        
    def setLineMedianKernel(self, value):
        self._lineMedianKernel = value
        
        # refresh the line profile plot
        self._on_slider_changed(value=None)

    def _on_slider_changed(self, value : int = None):
        """Respond to user changing the "Line Scan" slider.
        
        Args:
            value: The line profile number
        """
        if self._ba is None:
            return
        
        if value is None:
            value = self._currentLineNumber
        else:
            self._currentLineNumber = value

        logger.info(f'value:{value}')

        xScale = self._ba._abf.tifHeader['secondsPerLine']

        #secondsValue = value * self._kymographAnalysis._secondsPerLine
        secondsValue = value * xScale

        # update verical lines on top of plots
        for line in self._sliceLinesList:
            line.setValue(secondsValue)

        # update one line profile plot
        lineProfile, left_pnt, right_pnt = \
            self._kymographAnalysis._getFitLineProfile(value, self._lineMedianKernel)
        #print('    left:', left, 'right:', right)

        # logger.info(f'  len(lineProfile):{len(lineProfile)} \
        #             left_pnt:{left_pnt} \
        #             right_pnt:{right_pnt}')

        # +1 because np.arange() does NOT include the last point
        #xData = np.arange(0, self._kymographAnalysis.pointsPerLineScan()+1)
        pointsPerLineScan = self._kymographAnalysis.pointsPerLineScan()
        umPerPixel = self._kymographAnalysis._umPerPixel
        logger.info(f'pointsPerLineScan:{pointsPerLineScan} umPerPixel:{umPerPixel}')

        xData = np.arange(0, pointsPerLineScan)
        xData = np.multiply(xData, umPerPixel)
        if len(xData) != len(lineProfile):
            logger.error(f'xData: {len(xData)}')
            logger.error(f'lineProfile: {len(lineProfile)}')
        self.lineIntensityPlot.setData(xData, lineProfile)

        # if the fit fails we get left/right as nan
        if np.isnan(left_pnt) or np.isnan(right_pnt):
            xLeftRightData = [np.nan, np.nan]
            yLeftRightData = [np.nan, np.nan]
        else:
            left_um = self._kymographAnalysis.pnt2um(left_pnt)
            right_um = self._kymographAnalysis.pnt2um(right_pnt)
            xLeftRightData = [left_um, right_um] 
            #print(xLeftRightData)
            yLeftRightData = [lineProfile[left_pnt], lineProfile[right_pnt]] 
        self.leftRightPlot.setData(xLeftRightData, yLeftRightData)

    def _initGui(self):
        self.setWindowTitle('Kymograph Analysis')
        
        self._sliceLinesList = []
        
        vBoxLayout = QtWidgets.QVBoxLayout()

        # new oct 5 rosie
        #
        # kymograph widget from main interface
        self._kymWidgetMain = sanpy.interface.kymographWidget(self._ba) # will handle na is None
        self._kymWidgetMain.signalKymographRoiChanged.connect(self._slot_roi_changed2)
        self._kymWidgetMain.signalLineSliderChanged.connect(self._on_slider_changed)
        vBoxLayout.addWidget(self._kymWidgetMain)

        #
        # control bar
        hBoxLayoutControls = QtWidgets.QHBoxLayout()

        lineWidthLabel = QtWidgets.QLabel('Line Width (pixels)')
        hBoxLayoutControls.addWidget(lineWidthLabel)

        lineWidthSpinbox = QtWidgets.QSpinBox()
        lineWidthSpinbox.setMinimum(1)
        if self._kymographAnalysis is not None:
            lineWidthSpinbox.setValue(self._kymographAnalysis.getLineWidth())
        else:
            lineWidthSpinbox.setValue(1)
        lineWidthSpinbox.valueChanged.connect(self.setLineWidth)  # triggers as user types e.g. (22)
        hBoxLayoutControls.addWidget(lineWidthSpinbox)

        # percent of max in line profile (for analysis)
        percentMaxLabel = QtWidgets.QLabel('Percent Of Max')
        hBoxLayoutControls.addWidget(percentMaxLabel)

        percentMaxSpinbox = QtWidgets.QDoubleSpinBox()
        percentMaxSpinbox.setSingleStep(0.1)
        percentMaxSpinbox.setMinimum(0.001)
        if self._kymographAnalysis is not None:
            percentMaxSpinbox.setValue(self._kymographAnalysis.getPercentOfMax())
        else:
            percentMaxSpinbox.setValue(10)
        percentMaxSpinbox.valueChanged.connect(self.setPercentMax)  # triggers as user types e.g. (22)
        hBoxLayoutControls.addWidget(percentMaxSpinbox)

        # median kernels for image and line profile
        medianLabel = QtWidgets.QLabel('Median Image')
        hBoxLayoutControls.addWidget(medianLabel)

        imageMedianKernelSpinbox = QtWidgets.QSpinBox()
        imageMedianKernelSpinbox.setMinimum(0)
        imageMedianKernelSpinbox.setValue(self._imageMedianKernel)
        imageMedianKernelSpinbox.valueChanged.connect(self.setImageMedianKernel)  # triggers as user types e.g. (22)
        hBoxLayoutControls.addWidget(imageMedianKernelSpinbox)

        medianLabel2 = QtWidgets.QLabel('Line')
        hBoxLayoutControls.addWidget(medianLabel2)

        aSpinbox = QtWidgets.QSpinBox()
        aSpinbox.setMinimum(0)
        aSpinbox.setValue(self._lineMedianKernel)
        aSpinbox.valueChanged.connect(self.setLineMedianKernel)  # triggers as user types e.g. (22)
        hBoxLayoutControls.addWidget(aSpinbox)

        # buttons
        # buttonName = 'Reset ROI'
        # aButton = QtWidgets.QPushButton(buttonName)
        # aButton.clicked.connect(lambda state, name=buttonName: self._buttonCallback(name))
        # hBoxLayoutControls.addWidget(aButton)

        # buttonName = 'Rot -90'
        # aButton = QtWidgets.QPushButton(buttonName)
        # aButton.setEnabled(False)
        # aButton.clicked.connect(lambda state, name=buttonName: self._buttonCallback(name))
        # hBoxLayoutControls.addWidget(aButton)

        # keep reference to button so we can change color during analysis
        buttonName = 'Analyze'
        self._analyzeButton = QtWidgets.QPushButton(buttonName)
        self._analyzeButton.setStyleSheet("background-color : #22AA22")
        self._analyzeButton.clicked.connect(lambda state, name=buttonName: self._buttonCallback(name))
        hBoxLayoutControls.addWidget(self._analyzeButton)

        buttonName = 'Save'
        aButton = QtWidgets.QPushButton(buttonName)
        aButton.clicked.connect(lambda state, name=buttonName: self._buttonCallback(name))
        hBoxLayoutControls.addWidget(aButton)

        # 2nd row of controls
        hBoxLayoutControls2 = QtWidgets.QHBoxLayout()

        checkboxName = 'Line Profile'
        aCheckBox = QtWidgets.QCheckBox(checkboxName)
        aCheckBox.setChecked(True)
        aCheckBox.stateChanged.connect(lambda state, name=checkboxName: self._checkboxCallback(state,name))
        hBoxLayoutControls2.addWidget(aCheckBox)

        checkboxName = 'Diameter'
        aCheckBox = QtWidgets.QCheckBox(checkboxName)
        aCheckBox.setChecked(True)
        aCheckBox.stateChanged.connect(lambda state, name=checkboxName: self._checkboxCallback(state,name))
        hBoxLayoutControls2.addWidget(aCheckBox)

        checkboxName = 'Sum Line'
        aCheckBox = QtWidgets.QCheckBox(checkboxName)
        aCheckBox.setChecked(True)
        aCheckBox.stateChanged.connect(lambda state, name=checkboxName: self._checkboxCallback(state,name))
        hBoxLayoutControls2.addWidget(aCheckBox)

        checkboxName = 'Fit On Image'
        aCheckBox = QtWidgets.QCheckBox(checkboxName)
        aCheckBox.setChecked(self._fitIsVivible)
        aCheckBox.stateChanged.connect(lambda state, name=checkboxName: self._checkboxCallback(state,name))
        hBoxLayoutControls2.addWidget(aCheckBox)

        # 3rd row of controls
        # hBoxLayoutControls3 = QtWidgets.QHBoxLayout()

        # percentMaxLabel = QtWidgets.QLabel('y: um/pixel')
        # hBoxLayoutControls3.addWidget(percentMaxLabel)

        # percentMaxLabel = QtWidgets.QLabel('x: sec/line')
        # hBoxLayoutControls3.addWidget(percentMaxLabel)

        # add 2 rows of controls
        vBoxLayout.addLayout(hBoxLayoutControls)
        vBoxLayout.addLayout(hBoxLayoutControls2)
        # vBoxLayout.addLayout(hBoxLayoutControls3)

        # 1) kymograph image
        if 0:
            self.kymographWindow = pg.PlotWidget()
            self.kymographWindow.setLabel('left', 'Line Scan', units='')
            #self.kymographWindow.setLabel('bottom', 'Time', units='')

            self.kymographImage = pg.ImageItem(self._kymographAnalysis.getImage())
            # setRect(x, y, w, h)
            imageRect = self._kymographAnalysis.getImageRect()  # (x,y,w,h)
            #pos, size = self._kymographAnalysis.getFullRectRoi()  # (x,y,w,h)
            self.kymographImage.setRect(imageRect[0], imageRect[1], imageRect[2], imageRect[3])
            #self.kymographImage.setRect(pos[0], pos[1], size[0], size[1])
            self.kymographWindow.addItem(self.kymographImage)

            # vertical line to show "Line Profile"
            sliceLine = pg.InfiniteLine(pos=0, angle=90)
            self._sliceLinesList.append(sliceLine) # keep a list of vertical slice lines so we can update all at once
            self.kymographWindow.addItem(sliceLine)

            # rectangulaar roi over image
            #_penWidth = self._kymographAnalysis.getLineWidth()
            _penWidth = 1
            rectPen = pg.mkPen('r', width=_penWidth)
            pos = self._kymographAnalysis.getPosRoi()
            size = self._kymographAnalysis.getSizeRoi()
            _imageRect = self._kymographAnalysis.getImageRect()   # (x,y,w,h)
            maxBounds = QtCore.QRectF(_imageRect[0], _imageRect[1],
                                _imageRect[2], _imageRect[3])
            #print('  init roi pos:', pos)
            #print('  init roi size:', size)
            #print('  init roi pos:', maxBounds)
            self._rectRoi = pg.ROI(pos=pos,
                                size=size, maxBounds=maxBounds,
                                pen=rectPen)
            self._rectRoi.addScaleHandle(pos=(1,0), center=(0,1))
            self._rectRoi.sigRegionChangeFinished.connect(self._slot_roi_changed)  # remove
            self.kymographWindow.addItem(self._rectRoi)

            vBoxLayout.addWidget(self.kymographWindow)

        #
        # 1.5) slider to step through "Line Profile"
        if 0:
            self.profileSlider = QtWidgets.QSlider(QtCore.Qt.Orientation.Horizontal)
            self.profileSlider.setMinimum(0)
            self.profileSlider.setMaximum(self._kymographAnalysis.numLineScans())
            self.profileSlider.valueChanged.connect(self._on_slider_changed)
            vBoxLayout.addWidget(self.profileSlider)

        #
        # plot of just one line intensity
        self.lineIntensityPlotItem = pg.PlotWidget()
        self.lineIntensityPlotItem.setLabel('left', 'Intensity', units='')
        self.lineIntensityPlotItem.setLabel('bottom', 'Points', units='')
        self.lineIntensityPlot = self.lineIntensityPlotItem.plot(name='lineIntensityPlot')
        if self._kymographAnalysis is not None:
            xPlot = np.arange(0, self._kymographAnalysis.pointsPerLineScan())
        else:
            xPlot = np.arange(0,0)
        yPlot = xPlot * np.nan
        self.lineIntensityPlot.setData(xPlot, yPlot, connect='finite')  # fill with nan
        # single point of left/right
        self.leftRightPlot = self.lineIntensityPlotItem.plot(name='leftRightPlot')

        vBoxLayout.addWidget(self.lineIntensityPlotItem)

        #
        # 3) dimaeter of each line scan
        self.diameterPlotItem = pg.PlotWidget()
        self.diameterPlotItem.setLabel('left', 'Diameter', units='')
        self.diameterPlot = self.diameterPlotItem.plot(name='diameterPlot')
        if self._kymographAnalysis is not None:
            xPlot = np.arange(0, self._kymographAnalysis.numLineScans())
        else:
            xPlot = np.arange(0,0)
        yPlot = xPlot * np.nan
        self.diameterPlot.setData(xPlot, yPlot, connect='finite')  # fill with nan
        # link x-axis with kymograph PlotWidget
        #self.diameterPlotItem.setXLink(self.kymographWindow)
        # link to kymographWidget plot of the image
        self.diameterPlotItem.setXLink(self._kymWidgetMain.kymographPlot)

        # vertical line to show "Line Profile"
        sliceLine = pg.InfiniteLine(pos=0, angle=90)
        self._sliceLinesList.append(sliceLine) # keep a list of vertical slice lines so we can update all at once
        self.diameterPlotItem.addItem(sliceLine)

        vBoxLayout.addWidget(self.diameterPlotItem)

        # 4) sum intensity of each line scan
        self.sumIntensityPlotItem = pg.PlotWidget()
        self.sumIntensityPlotItem.setLabel('left', 'Sum Intensity', units='')
        self.sumIntensityPlot = self.sumIntensityPlotItem.plot(name='sumIntensityPlot')
        # TODO: fix
        if self._ba is not None:
            _recordingDur = self._ba.recordingDur
        else:
            _recordingDur = 0
        #xPlot = np.arange(0, self._kymographAnalysis.numLineScans())
        xPlot = np.arange(0, _recordingDur)
        yPlot = xPlot * np.nan
        self.sumIntensityPlot.setData(xPlot, yPlot, connect='finite')  # fill with nan
        # link x-axis with kymograph PlotWidget
        #self.sumIntensityPlotItem.setXLink(self.kymographWindow)
        # link to kymographWidget plot of the image
        self.sumIntensityPlotItem.setXLink(self._kymWidgetMain.kymographPlot)

        # vertical line to show "Line Profile"
        sliceLine = pg.InfiniteLine(pos=0, angle=90)
        self._sliceLinesList.append(sliceLine) # keep a list of vertical slice lines so we can update all at once
        self.sumIntensityPlotItem.addItem(sliceLine)

        vBoxLayout.addWidget(self.sumIntensityPlotItem)

        # finalize
        self.setLayout(vBoxLayout)

    def old_replot(self,path):

        logger.info('')
        
        self._kymographAnalysis = sanpy.kymAnalysis(path)
        
        if 0:
            # clear the image from the panel
            #self.kymographWindow.clear()
            
            self.kymographImage = pg.ImageItem(self._kymographAnalysis.getImage())
            # setRect(x, y, w, h)
            imageRect = self._kymographAnalysis.getImageRect()  # (x,y,w,h)
            self.kymographImage.setRect(imageRect[0], imageRect[1], imageRect[2], imageRect[3])
            self.kymographWindow.addItem(self.kymographImage)

            # update roi
            _imageRect = self._kymographAnalysis.getImageRect()   # (x,y,w,h)
            maxBounds = QtCore.QRectF(_imageRect[0], _imageRect[1],
                                _imageRect[2], _imageRect[3])
            self._rectRoi.maxBounds = maxBounds
            pos = self._kymographAnalysis.getPosRoi()
            size = self._kymographAnalysis.getSizeRoi()
            print(f'  pos:{pos}, size:{size}')
            self._rectRoi.setPos(pos)
            self._rectRoi.setSize(size)
            self._rectRoi.stateChanged()

        # update slider
        #self.profileSlider.setMaximum(self._kymographAnalysis.numLineScans())

        #
        #self._on_slider_changed(0)

        self.refreshSumLinePlot()
        self.refreshDiameterPlot()

    def refreshSumLinePlot(self):
        logger.info('')
        if self._ba is None:
            return
        numLineScans = self._kymographAnalysis.numLineScans()
        #xPlot = self._kymographAnalysis.getResults('time_ms')
        
        # xPlot = self._kymographAnalysis.getTimeArray()  # always the same
        # yPlot = self._kymographAnalysis.getResults('sumintensity')

        xPlot = self._ba.fileLoader.sweepX
        yPlot = self._ba.fileLoader.sweepY

        self.sumIntensityPlot.setData(xPlot, yPlot, connect='finite')

    def refreshDiameterPlot(self):
        logger.info('')
        if self._ba is None:
            return
        #numLineScans = self._kymographAnalysis.numLineScans()
        #xPlot = self._kymographAnalysis.getResults('time_ms')
        #xPlot = self._kymographAnalysis.getTimeArray()  # always the same
        xPlot = self._ba.fileLoader.sweepX
        yDiam_um = self._kymographAnalysis.getResults('diameter_um')

        self.diameterPlot.setData(xPlot, yDiam_um, connect='finite')

        # show start/stop of fit in main kymograph
        left_pnt = self._kymographAnalysis.getResults('left_pnt')
        right_pnt = self._kymographAnalysis.getResults('right_pnt')
        self._kymWidgetMain.updateLeftRightFit(left_pnt, right_pnt,
                                            visible=self._fitIsVivible)

if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)

    #app.setStyleSheet(qdarkstyle.load_stylesheet(qt_api=os.environ['PYQTGRAPH_QT_LIB']))
    #app.setStyleSheet(qdarkstyle.load_stylesheet())

    path = '/Users/cudmore/data/kym-example/control 2.5_0012.tif.frames/control 2.5_0012.tif'
    #path = '/Users/cudmore/data/sa-node-ca-video/HighK-aligned-8bit_ch1.tif'
    
    path = '/Users/cudmore/data/rosie/Raw data for contraction analysis/Female Old/filter median 1 C2-0-255 Cell 2 CTRL  2_5_21 female wt old.tif'
    
    ba = sanpy.bAnalysis(path)

    kw = kymographPlugin2(ba)
    kw.show()

    #print(kw._kymographAnalysis._getHeader())

    sys.exit(app.exec_())


"""
Dec 2022, NOT USED BY ANY OTHER CODE ?

Utilities to define analysis results.

These are keys in bAnalysis_ spike dict and columns in output reports
"""

import json
import os
from typing import Dict, List, Tuple

import numpy as np  # needed to convert np types to JSON types in save
import pandas as pd
import sanpy

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

defaultVal = float('nan')

# each key in analysisResultDict needs to have the same dict
def getDefaultDict():
    defaultDict = {
        'type': '',  # like: int, float, boolean, list
        'default': '',  # default value, can be 0, None, NaN, ...
        'units': '',  # real world units like point, mV, dvdt
        'depends on detection': '',  # organize documentation and refer to bDetect keys
        'error': '',  # if this analysis results can trigger an error
        'description': '',  # long description for documentation
        }
    return defaultDict.copy()

analysisResultDict = {}

key = 'analysisDate'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'str'
analysisResultDict[key]['default'] = ''
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['description'] = 'Date of analysis in yyyymmdd format.'

key = 'analysisTime'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'str'
analysisResultDict[key]['default'] = ''
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['description'] = 'Time of analysis in hh:mm:ss 24 hours format.'

key = 'modDate'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'str'
analysisResultDict[key]['default'] = ''
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['description'] = 'Modification date if AP is modified after detection.'

key = 'modTime'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'str'
analysisResultDict[key]['default'] = ''
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['description'] = 'Modification time if AP is modified after detection.'

key = 'analysisVersion'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'str'
analysisResultDict[key]['default'] = ''
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['description'] = 'Analysis version when analysis was run. See sanpy.analysisVersion'

key = 'interfaceVersion'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'str'
analysisResultDict[key]['default'] = ''
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['description'] = 'Interface version string when analysis was run. See sanpy.interfaceVersion'

key = 'file'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'str'
analysisResultDict[key]['default'] = ''
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['description'] = 'Name of raw data file analyzed'

key = 'detectionType'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = ''
analysisResultDict[key]['default'] = None
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['description'] = 'Type of detection, either vm or dvdt. See enum sanpy.bDetection.detectionTypes'

key = 'cellType'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'str'
analysisResultDict[key]['default'] = ''
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['description'] = 'User specified cell type'

key = 'sex'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'str'
analysisResultDict[key]['default'] = ''
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['description'] = 'User specified sex'

key = 'condition'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'str'
analysisResultDict[key]['default'] = ''
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['description'] = 'User specified condition'

#
# start real analysis results
key = 'sweep'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'int'
analysisResultDict[key]['default'] = 0  # todo: not sure
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['description'] = 'Sweep number of analyzed sweep. Zero based.'

key = 'epoch'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'int'
analysisResultDict[key]['default'] = np.nan  # todo: not sure
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['description'] = 'Stimulus epoch number the spike occured in. Zero based.'

key = 'epochLevel'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = np.nan  # todo: not sure
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['description'] = 'Epoch level (DAC) stimulus during the spike.'

key = 'sweepSpikeNumber'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'int'
analysisResultDict[key]['default'] = None  # todo: not sure
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['description'] = 'Spike number within the sweep. Zero based.'

key = 'spikeNumber'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'int'
analysisResultDict[key]['default'] = None  # todo: not sure
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['description'] = 'Spike number across all sweeps. Zero based.'

key = 'include'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'bool'
analysisResultDict[key]['default'] = True
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['description'] = 'Boolean indication include or not. Can be set by user/programmatically  after analysis.'

key = 'userType'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'int'
analysisResultDict[key]['default'] = 0
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['description'] = 'Integer indication user type. Can be set by user/programmatically  after analysis.'


key = 'errors'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'list'
analysisResultDict[key]['default'] = []
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['description'] = 'List of dictionary to hold detection errors for this spike'

#
# detection parameters
key = 'dvdtThreshold'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'dvdt'
analysisResultDict[key]['depends on detection'] = 'dvdtThreshold'
analysisResultDict[key]['description'] = 'AP Threshold in derivative dv/dt'

key = 'mvThreshold'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'mV'
analysisResultDict[key]['depends on detection'] = 'mvThreshold'
analysisResultDict[key]['description'] = 'AP Threshold in primary recording mV'

# todo: what about solay ...
key = 'medianFilter'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'int'
analysisResultDict[key]['default'] = 0
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['depends on detection'] = 'medianFilter'
analysisResultDict[key]['description'] = 'Median filter to generate filtered vm and dvdt. Value 0 indicates no filter.'

key = 'halfHeights'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'list'
analysisResultDict[key]['default'] = []
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['depends on detection'] = 'halfHeights'
analysisResultDict[key]['description'] = 'List of int to specify half-heights like [10, 20, 50, 80, 90].'

#
# actual value coming out of detection !!!!
key = 'thresholdPnt'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'int'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'point'
analysisResultDict[key]['description'] = 'AP threshold point'

key = 'thresholdSec'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'sec'
analysisResultDict[key]['description'] = 'AP threshold seconds'

key = 'thresholdVal'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'mV'  # TODO: will be pA for voltage-clamp
analysisResultDict[key]['description'] = 'Value of Vm at AP threshold point.'

key = 'thresholdVal_dvdt'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'dvdt'  # TODO: will be pA for voltage-clamp
analysisResultDict[key]['description'] = 'Value of dvdt at AP threshold point.'

key = 'dacCommand'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'mV'  # TODO: will be pA for voltage-clamp
analysisResultDict[key]['description'] = 'Value of DAC command at AP threshold point.'

key = 'peakPnt'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'int'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'point'
analysisResultDict[key]['depends on detection'] = '(onlyPeaksAbove_mV, peakWindow_ms)'
analysisResultDict[key]['description'] = 'AP peak point.'

key = 'peakSec'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'sec'
analysisResultDict[key]['description'] = 'AP peak seconds.'

key = 'peakVal'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'mV' # voltage-clamp'
analysisResultDict[key]['description'] = 'Value of Vm at AP peak point.'

key = 'peakHeight'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'mV'  # voltage-clamp
analysisResultDict[key]['description'] = 'Difference between peakVal minus thresholdVal.'

key = 'timeToPeak_ms'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'ms' # voltage-clamp'
analysisResultDict[key]['description'] = 'Time to peak (ms) after TOP.'

key = 'preMinPnt'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'int'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'point'  # voltage-clamp
analysisResultDict[key]['depends on detection'] = 'mdp_ms'
analysisResultDict[key]['description'] = 'Minimum before an AP taken from predefined window.'

key = 'preMinVal'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'mV'  # voltage-clamp
analysisResultDict[key]['description'] = 'Minimum before an AP taken from predefined window.'

key = 'preLinearFitPnt0'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'int'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'point'  # voltage-clamp
analysisResultDict[key]['description'] = 'Point where pre linear fit starts. Used for EDD Rate'

key = 'preLinearFitPnt1'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'int'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'point'  # voltage-clamp
analysisResultDict[key]['description'] = 'Point where pre linear fit stops. Used for EDD Rate'

key = 'earlyDiastolicDuration_ms'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'ms'  # voltage-clamp
analysisResultDict[key]['description'] = 'Time (ms) between start/stop of EDD.'

key = 'preLinearFitVal0'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'mv'  # voltage-clamp
analysisResultDict[key]['description'] = ''

key = 'preLinearFitVal1'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'mv'
analysisResultDict[key]['description'] = ''

key = 'earlyDiastolicDurationRate'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'mv/S'
analysisResultDict[key]['description'] = 'Early diastolic duration rate, the slope of the linear fit between start/stop of EDD.'

key = 'lateDiastolicDuration'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['description'] = 'Depreciated'

key = 'preSpike_dvdt_max_pnt'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'int'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'point'
analysisResultDict[key]['description'] = 'Point corresponding to peak in dv/dt before an AP.'

key = 'preSpike_dvdt_max_val'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'mV'
analysisResultDict[key]['description'] = 'Value of Vm at peak of dv/dt before an AP.'

key = 'preSpike_dvdt_max_val2'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'dv/dt'
analysisResultDict[key]['description'] = 'Value of dv/dt at peak of dv/dt before an AP.'

key = 'postSpike_dvdt_min_pnt'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'int'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'point'
analysisResultDict[key]['depends on detection'] = 'dvdtPostWindow_ms'
analysisResultDict[key]['description'] = 'Point corresponding to min in dv/dt after an AP.'

key = 'postSpike_dvdt_min_val'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'mV'
analysisResultDict[key]['description'] = 'Value of Vm at minimum of dv/dt after an AP.'

key = 'postSpike_dvdt_min_val2'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'dvdt'
analysisResultDict[key]['description'] = 'Value of dv/dt at minimum of dv/dt after an AP.'

key = 'isi_pnts'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'int'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'point'
analysisResultDict[key]['depends on detection'] = 'refractory_ms'
analysisResultDict[key]['description'] = 'Inter-Spike-Interval (points) with respect to previous AP.'

key = 'isi_ms'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'ms'
analysisResultDict[key]['description'] = 'Inter-Spike-Interval (ms) with respect to previous AP.'

key = 'spikeFreq_hz'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'Hz'
analysisResultDict[key]['description'] = 'AP frequency with respect to previous AP.'

key = 'cycleLength_pnts'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'int'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'point'
analysisResultDict[key]['description'] = 'Points between APs with respect to previous AP.'

key = 'cycleLength_ms'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'int'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'point'
analysisResultDict[key]['description'] = 'Time (ms) between APs with respect to previous AP.'

key = 'diastolicDuration_ms'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'float'
analysisResultDict[key]['default'] = defaultVal
analysisResultDict[key]['units'] = 'ms'
analysisResultDict[key]['description'] = 'Time (ms) between minimum before AP (preMinPnt) and AP time (thresholdPnt).'

key = 'widths'
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]['type'] = 'list'
analysisResultDict[key]['default'] = []
analysisResultDict[key]['units'] = ''
analysisResultDict[key]['description'] = 'A list of dict to hold half-height information for each half-height in detection halfHeights.'

# need to define keys width_<i> from halfHeights list [10, 20, 50, 80, 90]
for i in [10, 20, 50, 80, 90]:
    key = 'widths_' + str(i)
    analysisResultDict[key] = getDefaultDict()
    analysisResultDict[key]['type'] = 'int'
    analysisResultDict[key]['default'] = defaultVal
    analysisResultDict[key]['units'] = 'percent'
    analysisResultDict[key]['depends on detection'] = 'halfWidthWindow_ms'
    analysisResultDict[key]['description'] = f'Width (ms) at half-height {i} %.'

# now we have (AP threshold, AP peak), derive stats about each AP

def printDocs():
    """
    Print out human readable detection parameters and convert to markdown table.

    Requires:
        pip install tabulate

    See: bDetection.printDocs()
    """
    import pandas as pd
    from datetime import datetime

    logger.info('Ensure there are no errors')

    dictList = []
    for k, v in analysisResultDict.items():
        # iterating on getDefaultDict() to ensure all code above has valid k/v pairs
        #lineStr = k + '\t'
        oneDict = {
            'Name': k,
        }
        for k2 in getDefaultDict().keys():
            #print(f'  {k}: {k2}: {v[k2]}')
            #lineStr += f'{v[k2]}' + '\t'
            oneDict[k2] = v[k2]
        #
        #print(lineStr)

        dictList.append(oneDict)

        # check that k is in headerDefaultDict
        for k3 in v:
            if not k3 in getDefaultDict().keys():
                logger.error(f'Found extra key "{k}" in "analysisResultDict"')

    #
    df = pd.DataFrame(dictList)
    
    if 1:
        # to markdown for mkdocs md file
        #str = df.to_markdown()
        str = df.to_html()
        myDate = datetime.today().strftime('%Y-%m-%d')
        print(f'Generated {myDate} with sanpy.analysisVersion {sanpy.analysisVersion}')
        print(str)

    if 0:
        path = '/Users/cudmore/Desktop/sanpy-analysis-results.csv'
        print('saving:', path)
        df.to_csv(path, index=False)

class NumpyEncoder(json.JSONEncoder):
    """ Special json encoder for numpy types """
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)

class analysisResultList:
    """Class encapsulating a list of analysisResultDict (one dict per spike).
    """
    def __init__(self):
        # one copy for entire list
        
        # TODO: put xxx in a function getAnalysisResltDict()
        self._dDict = analysisResultDict

        # list of analysisResultDict
        self._myList = []

        self._iterIdx = -1

    def setFromListDict(self, listOfDict : List[dict]):
        """Set analysis results from a list of dict.
        
        Used when loading sanpy.bAnalysis from h5 file.

        When we create self (during spike detect) we have a list of class analysisResult.
        When we save/load we have a list of dict.

        This is assuming we re-create self every time we do spike detection
        """
        
        # do not do this, we are an analysisResultList as a list of analysisResult
        # self._myList = listOfDict
        
        self._myList = []
        for oneDict in listOfDict:
            oneAnalysisResult = analysisResult(theDict=oneDict)
            self._myList.append(oneAnalysisResult)

    def analysisDate(self):
        if len(self) > 0:
            return self._myList[0]['analysisDate']
        else:
            return None

    def analysisTime(self):
        if len(self) > 0:
            return self._myList[0]['analysisTime']
        else:
            return None
        
    def old_save(self, saveBase):
        savePath = saveBase + '-analysis.json'

        analysisList = self.asList()

        #print(analysisList[0].print())
        print(self._myList[0])

        with open(savePath, 'w') as f:
            json.dump(analysisList, f, cls=NumpyEncoder, indent=4)

    def old_load(self, loadBase):
        loadPath = loadBase + '-analysis.json'

        if not os.path.isfile(loadPath):
            logger.error(f'Did not find file: {loadPath}')
            return

        with open(loadPath, 'r') as f:
            self._myList = json.load(f)

    def appendDefault(self):
        """Append a spike to analysis.
        
        Used in bAnalysis spike detection.
        """
        oneResult = analysisResult()
        self._myList.append(oneResult)

    def appendAnalysis(self, analysisResultList):
        for analysisResult in analysisResultList:
            # analysisResult is for one spike
            self._myList.append(analysisResult)

    def addAnalysisResult(self, theKey, theDefault=None):

        # go through list and add to each [i] dict
        for spike in self:
            spike.addNewKey(theKey, theDefault=theDefault)

    def asList(self):
        """
        Return underlying list.
        """
        #return [spike.asDict() for spike in self._myList]
        return [x.asDict() for x in self._myList]

    def asDataFrame(self):
        """
        Note: underlying _myList is a list of analysisResult
        """
        return pd.DataFrame(self.asList())

    def __getitem__(self, key):
        """
        Allow [] indexing with self[int].
        """
        try:
            #return self._dDict[key]['currentValue']
            return self._myList[key]
        except (IndexError) as e:
            logger.error(f'{e}')

    def __len__(self):
        """ Allow len() with len(this)
        """
        return len(self._myList)

    def __iter__(self):
        """
        Allow iteration with "for item in self"
        """
        return self

    def __next__(self):
        """
        Allow iteration with "for item in self"
        """
        self._iterIdx += 1
        if self._iterIdx >= len(self._myList):
            self._iterIdx = -1 # reset to initial value
            raise StopIteration
        else:
            return self._myList[self._iterIdx]

class analysisResult:
    def __init__(self, theDict=None):
        """Create an anlysis item (for one spike)
        
        Args:
            theDict: Pre-existing dict when we load form h5 file
        """
        # this is the raw definition of analysis results (See above)
        # pull from this to create key/value self.rDict
        #self._dDict = analysisResultDict
        defaultDict = analysisResultDict
        # this is simple key/value pairs as we will in detection
        self._rDict = {}
        for k,v in defaultDict.items():
            default = v['default']
            self._rDict[k] = default

        if theDict is not None:
            for k,v in theDict.items():
                self[k] = v  # calls __setitem__()

    # this was interfering with converting to DataFrame ???
    '''
    def __str__(self):
        printList = []
        for k,v in self._rDict.items():
            if isinstance(v, dict):
                for k2,v2 in v.items():
                    printList.append(f'  {k2} : {v2} {type(v2)}')
            else:
                printList.append(f'{k} : {v} {type(v)}')
        return '\n'.join(printList)
    '''

    def print(self):
        printList = []
        for k,v in self._rDict.items():
            if isinstance(v, list):
                for item in v:
                    for k2,v2 in item.items():
                        printList.append(f'  {k2} : {v2} {type(v2)}')
            else:
                printList.append(f'{k} : {v} {type(v)}')
        return '\n'.join(printList)

    def addNewKey(self, theKey, theDefault=None):
        """
        Add a new key to this spike.

        Returns: (bool) True if new key added, false if key already exists.
        """
        if theDefault is None:
            #theType = 'float'
            theDefault = float('nan')

        # check if key exists
        keyExists = theKey in self._rDict.keys()
        addedKey = False
        if keyExists:
             #key exists, don't modify
             #logger.warning(f'The key "{theKey}" already exists and has value "{self._rDict[theKey]}"')
             pass
        else:
            self._rDict[theKey] = theDefault
            addedKey = True

        #
        return addedKey

    def asDict(self):
        """
        Returns underlying dictionary
        """
        return self._rDict

    def __getitem__(self, key):
        # to mimic a dictionary
        ret = None
        try:
            #return self._dDict[key]['currentValue']
            ret = self._rDict[key]
        except (KeyError) as e:
            logger.error(f'Error getting key "{key}" exception:{e}')
            raise
        #
        return ret

    def __setitem__(self, key, value):
        # to mimic a dictionary
        try:
            #self._dDict[key]['currentValue'] = value
            self._rDict[key] = value
        except (KeyError) as e:
            logger.error(f'{e}')

    def items(self):
        # to mimic a dictionary
        return self._rDict.items()

    def keys(self):
        # to mimic a dictionary
        return self._rDict.keys()


def test():
    ar = analysisResultList()

    for k,v in ar.items():
        print(k,v)


    key = 'analysisVersion'
    print(f'key:{key} value:"{ar[key]}" type:{type(ar[key])}')

    key = 'errors'
    print(f'key:{key} value:"{ar[key]}" type:{type(ar[key])}')

def test2():
    # load abf
    path = 'data/19114000.abf'
    ba = sanpy.bAnalysis(path)

    # detect
    sweepNumber = 0
    detectionClass = ba.detectionClass
    detectionClass['verbose'] = True
    ba.spikeDetect2__(sweepNumber, detectionClass)

    '''
    printSpikeNum = 4
    print(f'== printing spike {printSpikeNum}')
    ba.printSpike(printSpikeNum)
    '''

    #ba.printErrors()

    sd = ba.getSpikeDictionaries()
    '''
    for idx, s in enumerate(sd):
        if idx == 2:
            print(s)
    '''

if __name__ == '__main__':
    #test()
    #test2()
    printDocs()
